import pandas as pd
from django.apps import apps as django_apps
from django_pandas.io import read_frame

from edc_appointment.models import AppointmentType

from ..utils import convert_dates_from_model


def convert_visit_code_to_float(df):
    """Convert visit_code to float using visit_code_sequence"""
    df["visit_code"] = df["visit_code"].astype(float)
    df["visit_code_sequence"] = df["visit_code_sequence"].astype(float)
    df["visit_code_sequence"] = df["visit_code_sequence"].apply(
        lambda x: x / 10.0 if x > 0.0 else 0.0
    )
    df["visit_code"] = df["visit_code"] + df["visit_code_sequence"]


def get_subject_visit(
    model: str,
    subject_identifiers: list[str] | None = None,
    normalize: bool | None = None,
    localize: bool | None = None,
) -> pd.DataFrame:
    """Read subject visit django model.

    Converts visit_code to a float of visit_code + visit_code_sequence,
    For example:
        1000.0, 1000.1, ...
    The original string visit_code is renamed as visit_code_str.

    Adds baseline and endline visit datetime and endline_visit_code
    """
    normalize = True if normalize is None else normalize
    localize = True if localize is None else localize
    model_cls = django_apps.get_model(model)

    values = [
        "id",
        "subject_identifier",
        "visit_code",
        "visit_code_sequence",
        "report_datetime",
        "site",
        "reason",
        "reason_unscheduled",
        "reason_unscheduled_other",
        "reason_missed",
        "reason_missed_other",
        "appointment",
        "appointment__appt_datetime",
        "appointment__appt_status",
        "appointment__appt_timing",
        "appointment__appt_type",
    ]
    if subject_identifiers:
        qs_subject_visit = model_cls.objects.values(*values).filter(
            subject_identifier__in=subject_identifiers
        )
    else:
        qs_subject_visit = model_cls.objects.values(*values).all()
    df = read_frame(qs_subject_visit, verbose=False)
    df = convert_dates_from_model(df, model_cls, normalize=normalize, localize=localize)
    df = df.rename(
        columns={
            "id": "subject_visit_id",
            "report_datetime": "visit_datetime",
            "site": "site_id",
            "appointment": "appointment_id",
            "appointment__appt_datetime": "appt_datetime",
            "appointment__appt_status": "appt_status",
            "appointment__appt_timing": "appt_timing",
            "appointment__appt_type": "appt_type",
        }
    )
    df["visit_code_str"] = df["visit_code"]
    df = df[
        [
            "subject_visit_id",
            "subject_identifier",
            "visit_code",
            "visit_code_sequence",
            "visit_datetime",
            "site_id",
            "visit_code_str",
            "reason",
            "reason_unscheduled",
            "reason_unscheduled_other",
            "reason_missed",
            "reason_missed_other",
            "appointment_id",
            "appt_datetime",
            "appt_status",
            "appt_timing",
            "appt_type",
        ]
    ]

    # map appt_type
    mapping = {obj.id: obj.name for obj in AppointmentType.objects.all().order_by("id")}
    df["appt_type"] = df["appt_type"].map(mapping)

    df["visit_datetime"] = df["visit_datetime"].apply(pd.to_datetime)

    convert_visit_code_to_float(df)

    df_baseline_visit = df.copy()
    df_baseline_visit = df_baseline_visit[(df_baseline_visit["visit_code"] == 1000.0)]
    df_baseline_visit = df_baseline_visit.rename(
        columns={"visit_datetime": "baseline_datetime"}
    )
    df_baseline_visit = df_baseline_visit[["subject_identifier", "baseline_datetime"]]

    df = df.merge(df_baseline_visit, on="subject_identifier", how="left")

    # get last visitcode and last visit datetime
    df_last = (
        df[~df.reason.isin(["missed", "Missed visit"])]
        .groupby("subject_identifier")
        .agg({"visit_code": "max", "visit_datetime": "max"})
    ).copy()
    df_last = df_last.rename(
        columns={
            "visit_code": "endline_visit_code",
            "visit_datetime": "endline_visit_datetime",
        }
    )
    df_last["endline_visit_code_str"] = (
        df_last["endline_visit_code"].astype("int64").apply(lambda x: str(x))
    )
    df_last = df_last.reset_index()
    df = df.merge(df_last, on="subject_identifier", how="left")

    df["followup_days"] = (df.visit_datetime - df.baseline_datetime).dt.days

    # get next visitcode and next visit datetime, if there is one
    # df_next = get_appointment_df()
    # df_next = (
    #     df_next[df.appt_status == NEW_APPT]
    #     .groupby("subject_identifier")
    #     .agg({"visit_code": "min", "visit_datetime": "min"})
    # )

    df = df.sort_values(by=["subject_identifier", "visit_code"])
    df = df.reset_index(drop=True)
    return df



from django.core.management import BaseCommand
from tqdm import tqdm

from edc_appointment.models import Appointment
from edc_appointment.utils import update_appt_status


class Command(BaseCommand):
    help = "Update appointment status for all appointments"

    def handle(self, *args, **options) -> None:
        appointments = Appointment.objects.all().order_by("subject_identifier")
        total = appointments.count()
        for appointment in tqdm(appointments, total=total):
            update_appt_status(appointment, save=True)
        print("\n\nDone")



from __future__ import annotations

import sys
from collections.abc import Callable
from copy import deepcopy
from warnings import warn

from django.apps import apps as django_apps
from django.conf import settings
from django.utils.module_loading import import_module, module_has_submodule

from .auth_objects import default_pii_models, get_default_groups, get_default_roles
from .constants import (
    CUSTOM_PERMISSIONS_TUPLES_KEY,
    GROUPS_KEY,
    PII_MODELS_KEY,
    POST_UPDATE_FUNCS_KEY,
    PRE_UPDATE_FUNCS_KEY,
    ROLES_KEY,
    UPDATE_GROUPS_KEY,
    UPDATE_ROLES_KEY,
)


class AlreadyRegistered(Exception):  # noqa: N818
    pass


class InvalidGroup(Exception):  # noqa: N818
    pass


class InvalidRole(Exception):  # noqa: N818
    pass


class RoleAlreadyExists(Exception):  # noqa: N818
    pass


class GroupAlreadyExists(Exception):  # noqa: N818
    pass


class PiiModelAlreadyExists(Exception):  # noqa: N818
    pass


class SiteAuthError(Exception):
    pass


def is_view_codename(codename):
    return (
        "view_" in codename
        or "view_historical" in codename
        or "nav_" in codename
        or "navbar" in codename
        or "dashboard" in codename
    )


def view_only_wrapper(func):
    codenames = func()
    return [codename for codename in codenames if is_view_codename(codename)]


def convert_view_to_export_wrapper(codename_or_callables):
    codenames = []
    export_codenames = []
    for codename in codename_or_callables:
        try:
            codenames.extend(codename())
        except TypeError:
            codenames.append(codename)
    for codename in codenames:
        if is_view_codename(codename) and "historical" not in codename:
            try:
                django_apps.get_model(codename.replace("view_", ""))
            except LookupError:
                pass
            else:
                export_codenames.append(codename.replace("view_", "export_"))
    return export_codenames


def remove_delete_wrapper(codename_or_callables):
    codenames = []
    for codename in codename_or_callables:
        try:
            codenames.extend(codename())
        except TypeError:
            codenames.append(codename)
    return [c for c in codenames if "delete_" not in c]


class SiteAuths:
    """A global to hold the intended group and role data.

    Data will be used by `AuthUpdater`.
    """

    def __init__(self):
        self.registry = {}
        self.initialize()

    def initialize(self):
        self.registry = {
            GROUPS_KEY: get_default_groups(),
            ROLES_KEY: get_default_roles(),
            UPDATE_GROUPS_KEY: {},
            UPDATE_ROLES_KEY: {},
            CUSTOM_PERMISSIONS_TUPLES_KEY: {},
            PRE_UPDATE_FUNCS_KEY: [],
            POST_UPDATE_FUNCS_KEY: [],
            PII_MODELS_KEY: default_pii_models,
        }

    def clear(self):
        self.registry = {
            GROUPS_KEY: {},
            ROLES_KEY: {},
            UPDATE_GROUPS_KEY: {},
            UPDATE_ROLES_KEY: {},
            CUSTOM_PERMISSIONS_TUPLES_KEY: {},
            PRE_UPDATE_FUNCS_KEY: [],
            POST_UPDATE_FUNCS_KEY: [],
            PII_MODELS_KEY: [],
        }

    def clear_values(self):
        registry = deepcopy(self.registry)
        self.registry = {
            GROUPS_KEY: {k: [] for k in registry.get(GROUPS_KEY)},
            ROLES_KEY: {k: [] for k in self.registry.get(ROLES_KEY)},
            UPDATE_GROUPS_KEY: {},
            UPDATE_ROLES_KEY: {},
            CUSTOM_PERMISSIONS_TUPLES_KEY: {},
            PRE_UPDATE_FUNCS_KEY: [],
            POST_UPDATE_FUNCS_KEY: [],
            PII_MODELS_KEY: [],
        }

    @property
    def edc_auth_skip_site_auths(self):
        return getattr(settings, "EDC_AUTH_SKIP_SITE_AUTHS", False)

    def add_pre_update_func(self, func):
        self.registry[PRE_UPDATE_FUNCS_KEY].append(func)

    def add_post_update_func(self, app_label: str, func: Callable):
        self.registry[POST_UPDATE_FUNCS_KEY].append((app_label, func))

    def add_pii_model(self, model_name):
        if model_name in self.registry[PII_MODELS_KEY]:
            raise PiiModelAlreadyExists(f"PII model already exists. Got {model_name}")
        self.registry[PII_MODELS_KEY].append(model_name)

    def add_groups(self, data: dict):
        for name, codenames in data.items():
            self.add_group(codenames, name=name)

    def add_roles(self, data: dict):
        for name, group_names in data.items():
            self.add_role(group_names, name=name)

    def add_group(
        self,
        *codenames_or_func,
        name=None,
        view_only=None,
        convert_to_export=None,
        no_delete=None,
    ):
        if name in self.registry[GROUPS_KEY]:
            raise GroupAlreadyExists(f"Group name already exists. Got {name}.")
        if no_delete:
            codenames_or_func = self.remove_delete_codenames(codenames_or_func)
        if view_only:
            codenames_or_func = self.get_view_only_codenames(codenames_or_func)
        if convert_to_export:
            codenames_or_func = self.convert_to_export_codenames(codenames_or_func)
        self.registry[GROUPS_KEY].update({name: codenames_or_func})

    def add_role(self, *group_names, name=None):
        if name in self.registry[ROLES_KEY]:
            raise RoleAlreadyExists(f"Role name already exists. Got {name}.")
        group_names = list(set(group_names))
        self.registry[ROLES_KEY].update({name: group_names})

    def update_group(
        self, *codenames_or_func, name=None, key=None, view_only=None, no_delete=None
    ) -> None:
        key = key or UPDATE_GROUPS_KEY
        if no_delete:
            codenames_or_func = self.remove_delete_codenames(codenames_or_func)
        if view_only:
            codenames_or_func = self.get_view_only_codenames(codenames_or_func)
        codenames_or_func = list(set(codenames_or_func))
        existing_codenames = deepcopy(self.registry[key].get(name)) or []
        try:
            existing_codenames = list(set(existing_codenames))
        except TypeError as e:
            raise TypeError(f"{e}. Got {name}") from e
        existing_codenames.extend(codenames_or_func)
        existing_codenames = list(set(existing_codenames))
        self.registry[key].update({name: existing_codenames})

    def update_role(self, *group_names, name=None, key=None) -> None:
        key = key or UPDATE_ROLES_KEY
        group_names = list(set(group_names))
        existing_group_names = [
            name for name in self.registry[key].get(name) or [] if name not in group_names
        ]
        existing_group_names.extend(group_names)
        self.registry[key].update({name: existing_group_names})

    def add_custom_permissions_tuples(
        self, model: str, codename_tuples: tuple[tuple[str, str], ...]
    ):
        try:
            self.registry[CUSTOM_PERMISSIONS_TUPLES_KEY][model]
        except KeyError:
            self.registry[CUSTOM_PERMISSIONS_TUPLES_KEY].update({model: []})
        for codename_tuple in codename_tuples:
            if codename_tuple not in self.registry[CUSTOM_PERMISSIONS_TUPLES_KEY][model]:
                self.registry[CUSTOM_PERMISSIONS_TUPLES_KEY][model].append(codename_tuple)

    @staticmethod
    def get_view_only_codenames(codenames):
        """Returns a list of view only codenames.

        If codename is a callable, wraps for a later call.

        Does not remove `edc_navbar`, 'nav_' or `edc_dashboard`
        codenames.
        """
        # callables = [lambda: view_only_wrapper(c) for c in codenames if callable(c)]
        callables = [lambda c=c: view_only_wrapper(c) for c in codenames if callable(c)]
        view_only_codenames = [
            codename
            for codename in codenames
            if not callable(codename) and is_view_codename(codename)
        ]
        view_only_codenames.extend(callables)
        return view_only_codenames

    @staticmethod
    def convert_to_export_codenames(codenames):
        """Returns a list of export only codenames by
        replacing `view` codenames with `export`.

        If codename is a callable, wraps for a later call.
        """
        export_codenames = []
        callables = [codename for codename in codenames if callable(codename)]
        codenames = [codename for codename in codenames if not callable(codename)]
        if callables:
            export_codenames.append(lambda: convert_view_to_export_wrapper(callables))
        if codenames:
            export_codenames.extend(convert_view_to_export_wrapper(codenames))
        return export_codenames

    @staticmethod
    def remove_delete_codenames(codenames):
        export_codenames = []
        callables = [codename for codename in codenames if callable(codename)]
        codenames = [codename for codename in codenames if not callable(codename)]
        if callables:
            export_codenames.append(lambda: remove_delete_wrapper(callables))
        if codenames:
            export_codenames.extend(remove_delete_wrapper(codenames))
        return export_codenames

    @property
    def roles(self):
        return self.registry[ROLES_KEY]

    @property
    def groups(self):
        return self.registry[GROUPS_KEY]

    @property
    def pii_models(self):
        return self.registry[PII_MODELS_KEY]

    @property
    def pre_update_funcs(self):
        return self.registry[PRE_UPDATE_FUNCS_KEY]

    @property
    def post_update_funcs(self) -> tuple[str, Callable]:
        return self.registry[POST_UPDATE_FUNCS_KEY]

    @property
    def custom_permissions_tuples(self):
        return self.registry[CUSTOM_PERMISSIONS_TUPLES_KEY]

    def verify_and_populate(
        self, app_name: str | None = None, warn_only: bool | None = None
    ) -> None:
        """Verifies that updates refer to existing group
        or roles names.

        * Updates data from `update_groups` -> `groups`
        * Updates data from `update_roles` -> `roles`
        """
        for name, codenames in self.registry[UPDATE_GROUPS_KEY].items():
            if name not in self.registry[GROUPS_KEY]:
                msg = (
                    f"Cannot update group. Group name does not exist. See app={app_name}"
                    f"update_groups['groups']={codenames}. Got {name}"
                )
                if warn_only:
                    warn(msg, stacklevel=2)
                else:
                    raise InvalidGroup(msg)
            self.update_group(*codenames, name=name, key=GROUPS_KEY)
        self.registry[UPDATE_GROUPS_KEY] = {}
        for name, group_names in self.registry[UPDATE_ROLES_KEY].items():
            if name not in self.registry[ROLES_KEY]:
                msg = (
                    f"Cannot update role. Role name does not exist. See app={app_name}. "
                    f"update_roles['groups']={group_names}. Got {name}"
                )
                if warn_only:
                    warn(msg, stacklevel=2)
                else:
                    raise InvalidRole(msg)
            self.update_role(*group_names, name=name, key=ROLES_KEY)
        self.registry[UPDATE_ROLES_KEY] = {}

    def autodiscover(self, module_name=None, verbose=True):
        """Autodiscovers in the auths.py file of any INSTALLED_APP."""
        if not self.edc_auth_skip_site_auths:
            before_import_registry = None
            module_name = module_name or "auths"
            writer = sys.stdout.write if verbose else lambda x: x
            writer(f" * checking for site {module_name} ...\n")
            for app_name in django_apps.app_configs:
                writer(f" * searching {app_name}           \r")
                try:
                    mod = import_module(app_name)
                    try:
                        before_import_registry = deepcopy(site_auths.registry)
                        import_module(f"{app_name}.{module_name}")
                        writer(f"   - registered '{module_name}' from '{app_name}'\n")
                    except ImportError as e:
                        site_auths.registry = before_import_registry
                        if module_has_submodule(mod, module_name):
                            raise SiteAuthError(str(e)) from e
                except ImportError:
                    pass
            self.verify_and_populate(app_name=app_name)


site_auths = SiteAuths()



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING

from edc_appointment.constants import MISSED_APPT

from .logic import Logic
from .rule_evaluator import RuleEvaluator

if TYPE_CHECKING:
    from edc_visit_tracking.model_mixins import VisitModelMixin as Base

    from ..model_mixins.creates import CreatesMetadataModelMixin
    from .predicate import PF, P

    class RelatedVisitModel(CreatesMetadataModelMixin, Base):
        pass


class RuleError(Exception):
    pass


class Rule:
    rule_evaluator_cls = RuleEvaluator
    logic_cls = Logic

    def __init__(
        self,
        predicate: P | PF | callable | str = None,
        consequence: str = None,
        alternative: str = None,
    ) -> None:
        self.predicate = predicate
        self.consequence = consequence
        self.alternative = alternative
        self.target_models: list[str] | None = None
        self.app_label: str | None = None  # set by metaclass
        self.group = None  # set by metaclass
        self.name: str | None = None  # set by metaclass
        self.source_model: str | None = None  # set by metaclass
        self.related_visit_model: str | None = None  # set by metaclass

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name='{self.name}', group='{self.group}')"

    def __str__(self) -> str:
        return f"{self.group}.{self.name}"

    def run(self, related_visit: RelatedVisitModel = None) -> dict[str, str] | None:
        """Returns a dictionary of {target_model: entry_status, ...} updated
        by running the rule for each target model given a visit.

        Skips run if `appointment.appt_timing` == MISSED_APPT
        """
        result = None
        if (
            self.related_visit_model
            and self.related_visit_model != related_visit._meta.label_lower
        ):
            raise RuleError(
                "Conflicting related visit model on rule. "
                f"Got {self.related_visit_model} != {related_visit._meta.label_lower}."
                "Try specifying the related visit model on RuleGroup.Meta explicitly. "
                f'For example, related_visit_model = "{related_visit._meta.label_lower}" '
                f"See {self}. "
            )

        if related_visit.appointment.appt_timing != MISSED_APPT:
            result = {}
            opts = {k: v for k, v in self.__dict__.items() if k.startswith != "_"}
            rule_evaluator = self.rule_evaluator_cls(
                related_visit=related_visit, logic=self.logic, **opts
            )
            entry_status = rule_evaluator.result
            for target_model in self.target_models:
                result.update({target_model: entry_status})
        return result

    @property
    def logic(self) -> Logic:
        return self.logic_cls(
            predicate=self.predicate,
            consequence=self.consequence,
            alternative=self.alternative,
        )

    @property
    def field_names(self) -> list[str]:
        field_names = []
        try:
            field_names = [self.predicate.attr]
        except AttributeError:
            with contextlib.suppress(AttributeError):
                field_names = self.predicate.attrs
        return field_names



import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype

from ..constants import COUNT_COLUMN, N_ONLY
from ..styler import Styler


class RowStatistics:
    """A class that calculates descriptive statistics for an
    indictor.
    """

    def __init__(
        self,
        colname: str = None,
        df_numerator: pd.DataFrame = None,
        df_denominator: pd.DataFrame = None,
        df_all: pd.DataFrame = None,
        coltotal: float | int | None = None,
        style: str | None = None,
        places: int | None = None,
    ):
        self.places = 2 if places is None else places
        self.style = style or N_ONLY

        # counts (6 columns)
        self.count = 0.0 if df_numerator.empty else len(df_numerator)
        self.total = len(df_all)
        self.coltotal = coltotal or len(df_denominator)
        self.rowtotal = self.count  # rowtotal or len(df_denominator)
        self.colprop = self.count / self.coltotal if self.count else 0.0
        self.rowprop = self.count / self.total if self.count else 0.0

        # numeric stats (9 columns)
        if colname and not df_numerator.empty and is_numeric_dtype(df_numerator[colname]):
            stats = df_numerator[colname].describe()
            self.mean = stats.loc["mean"]
            self.sd = stats.loc["std"]
            self.min = stats.loc["min"]
            self.max = stats.loc["max"]
            self.q25, self.q50, self.q75 = df_numerator[colname].quantile([0.25, 0.50, 0.75])
            stats = df_numerator[colname].agg(["mean", "sem"])
            self.ci95l = stats.loc["mean"] - 1.96 * stats.loc["sem"]
            self.ci95h = stats.loc["mean"] + 1.96 * stats.loc["sem"]
        else:
            (
                self.mean,
                self.sd,
                self.min,
                self.max,
                self.q25,
                self.q50,
                self.q75,
                self.ci95l,
                self.ci95h,
            ) = [np.nan] * 9

    def values_list(self) -> list:
        return list(self.as_dict().values())

    def labels(self) -> list:
        return list(self.as_dict().keys())

    def as_dict(self):
        return {
            COUNT_COLUMN: self.count,
            "coltotal": self.coltotal,
            "rowtotal": self.rowtotal,
            "total": self.total,
            "colprop": self.colprop,
            "rowprop": self.rowprop,
            "mean": self.mean,
            "sd": self.sd,
            "min": self.min,
            "max": self.max,
            "q25": self.q25,
            "q50": self.q50,
            "q75": self.q75,
            "ci95l": self.ci95l,
            "ci95h": self.ci95h,
        }

    def formatted_cell(self) -> str:
        return Styler(style=self.style, statistics=self, places=self.places).value

    def row(self):
        return [self.formatted_cell()] + self.values_list()



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "edc_lab_dashboard"
    verbose_name = "Edc Lab Dashboard"
    include_in_administration_section = False
    admin_site_name = "edc_lab_admin"



from django.urls import path

from .views import PdfIntermediateView, PrintPdfReportView

app_name = "edc_pdf_reports"

urlpatterns = [
    path(
        "pdf_report/intermediate/<app_label>/<model_name>/<pk>/",
        PdfIntermediateView.as_view(),
        name="pdf_intermediate_url",
    ),
    path(
        "pdf_report/<app_label>/<model_name>/<pk>/",
        PrintPdfReportView.as_view(),
        name="pdf_report_url",
    ),
    path(
        "pdf_report/<app_label>/<model_name>/",
        PrintPdfReportView.as_view(),
        name="pdf_report_url",
    ),
]



from typing import Any

from django.core.exceptions import ObjectDoesNotExist
from django.db.models.signals import m2m_changed, post_delete, post_save
from django.dispatch import receiver
from simple_history.signals import post_create_historical_record

from edc_constants.constants import CLOSED, NEW, OPEN
from edc_notification.site_notifications import site_notifications

from ..utils import reset_and_delete_action_item
from .action_item import ActionItem
from .action_model_mixin import ActionModelMixin, ActionNoManagersModelMixin


def update_or_create_action_item(instance: Any, created: bool, using: dict):
    action_item = None
    action_cls = instance.get_action_cls()
    if not instance.action_item:
        action_item = ActionItem.objects.using(using).get(
            action_identifier=instance.action_identifier,
            subject_identifier=instance.subject_identifier,
        )
        instance.action_item = action_item
    # instantiate action class
    action_cls(
        action_item=action_item or instance.action_item,
        subject_identifier=instance.subject_identifier,
        using=using,
    )
    if created and instance.action_item.status == NEW:
        instance.action_item.status = OPEN
        instance.action_item.save(using=using)
    update_action_item_reason(instance)


def update_action_item_reason(instance):
    action_item_reason = instance.get_action_item_reason()
    if action_item_reason:
        instance.action_item_reason = action_item_reason
        instance.save_base(update_fields=["action_item_reason"])


@receiver(m2m_changed, weak=False, dispatch_uid="update_action_item_reason_on_m2m_changed")
def update_action_item_reason_on_m2m_changed(action, instance, **kwargs):
    if "historical" not in instance._meta.label_lower and isinstance(
        instance, (ActionModelMixin, ActionNoManagersModelMixin)
    ):
        update_action_item_reason(instance)


@receiver(m2m_changed, weak=False, dispatch_uid="update_or_create_action_item_on_m2m_change")
def update_or_create_action_item_on_m2m_change(action, instance, using, **kwargs):
    if "historical" not in instance._meta.label_lower and isinstance(
        instance, (ActionModelMixin, ActionNoManagersModelMixin)
    ):
        update_or_create_action_item(instance, False, using)


@receiver(post_save, weak=False, dispatch_uid="update_or_create_action_item_on_post_save")
def update_or_create_action_item_on_post_save(
    sender, instance, raw, created, using, update_fields, **kwargs
):
    """Updates action item for a model using the ActionModelMixin.

    The update is done by instantiating the action class associated
    with this model's instance.
    """
    if not raw and not update_fields:
        try:
            instance.action_name  # noqa
            instance.action_item  # noqa
        except AttributeError as e:
            if "action_name" not in str(e) and "action_item" not in str(e):
                raise
        else:
            if "historical" not in instance._meta.label_lower and isinstance(
                instance, (ActionModelMixin, ActionNoManagersModelMixin)
            ):
                update_or_create_action_item(instance, created, using)


@receiver(post_delete, weak=False, dispatch_uid="action_on_reference_model_post_delete")
def action_on_reference_model_post_delete(sender, instance: Any, using, **kwargs):
    """Re-opens an action item when the action's reference
    model is deleted.

    Also removes any "next" actions.

    Recreates the next action if needed.
    """
    if not isinstance(instance, ActionItem):
        try:
            instance.get_action_cls()
        except AttributeError as e:
            if "get_action_cls" not in str(e):
                raise
        else:
            reset_and_delete_action_item(instance, using)
    elif isinstance(instance, ActionItem) and instance.parent_action_item:
        try:
            parent_reference_obj = instance.parent_reference_obj
        except ObjectDoesNotExist:
            pass
        else:
            parent_reference_obj.action_item.action_cls(
                action_item=instance.parent_reference_obj.action_item,
                subject_identifier=instance.subject_identifier,
                using=using,
            ).create_next_action_items()


@receiver(
    post_create_historical_record,
    weak=False,
    dispatch_uid="action_item_notification_on_post_create_historical_record",
)
def action_item_notification_on_post_create_historical_record(
    sender, instance, history_date, history_user, history_change_reason, **kwargs
):
    """Checks and processes any notifications for this model.

    Note, this is the post_create of the historical model.
    """
    if (
        site_notifications.loaded
        and instance._meta.label_lower == "edc_action_item.actionitem"
        and instance.status != CLOSED
    ):
        opts = dict(
            instance=instance,
            user=instance.user_modified or instance.user_created,
            history_date=history_date,
            history_user=history_user,
            history_change_reason=history_change_reason,
            fail_silently=True,
            **kwargs,
        )
        site_notifications.notify(**opts)



from django import forms

from ...schedule import Schedule
from ...subject_schedule import (
    NotOnScheduleError,
    NotOnScheduleForDateError,
    SubjectSchedule,
)
from ...utils import report_datetime_within_onschedule_offschedule_datetimes
from ...visit_schedule import VisitSchedule


class VisitScheduleCrfModelFormMixin:
    """A ModelForm mixin for CRFs to validate that the subject is
    onschedule at the time of the report.

    Required to be declared together with other Crf modelform mixins.

    See also CrfModelFormMixin
    """

    def clean(self):
        cleaned_data = super().clean()
        self.is_onschedule_or_raise()
        self.report_datetime_within_schedule_datetimes()
        return cleaned_data

    @property
    def visit_schedule_name(self) -> str:
        return self.related_visit.visit_schedule_name

    @property
    def schedule_name(self) -> str:
        return self.related_visit.schedule_name

    @property
    def visit_schedule(self) -> VisitSchedule:
        return self.related_visit.visit_schedule

    @property
    def schedule(self) -> Schedule:
        return self.related_visit.schedule

    def is_onschedule_or_raise(self) -> None:
        if self.report_datetime and self.related_visit:
            visit_schedule = self.visit_schedule
            schedule = self.schedule
            subject_schedule = SubjectSchedule(
                self.get_subject_identifier(),
                visit_schedule=visit_schedule,
                schedule=schedule,
            )
            try:
                subject_schedule.onschedule_or_raise(
                    report_datetime=self.report_datetime,
                    compare_as_datetimes=(
                        self._meta.model.offschedule_compare_dates_as_datetimes
                    ),
                )
            except (NotOnScheduleError, NotOnScheduleForDateError) as e:
                raise forms.ValidationError(str(e)) from e

    def report_datetime_within_schedule_datetimes(self) -> None:
        if self.report_datetime:
            report_datetime_within_onschedule_offschedule_datetimes(
                subject_identifier=self.get_subject_identifier(),
                report_datetime=self.report_datetime,
                visit_schedule_name=self.visit_schedule_name,
                schedule_name=self.schedule_name,
                exception_cls=forms.ValidationError,
            )



# from importlib.metadata import version
#
# __version__ = version("edc_fieldsets")
#
# from .fieldlist import Fieldlist, Insert, Remove
# from .fieldset import Fieldset
# from .fieldsets import FieldsetError, Fieldsets
# from .fieldsets_modeladmin_mixin import FieldsetsModelAdminMixin



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps

if TYPE_CHECKING:
    from edc_reportable.models import MolecularWeight


def molecular_weight_model_cls() -> type[MolecularWeight]:
    return django_apps.get_model("edc_reportable", "MolecularWeight")



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from edc_model_admin.history import SimpleHistoryAdmin
from edc_randomization.blinding import user_is_blinded_from_request

from ...admin_site import edc_pharmacy_admin
from ...forms import LotForm
from ...models import Lot
from ..model_admin_mixin import ModelAdminMixin
from ..remove_fields_for_blinded_users import remove_fields_for_blinded_users


@admin.register(Lot, site=edc_pharmacy_admin)
class LotAdmin(ModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Batches"
    change_form_title = "Pharmacy: Batch"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20

    form = LotForm

    fieldsets = (
        (
            None,
            {
                "fields": [
                    "lot_no",
                    "product",
                    "manufactured_date",
                    "processed_until_date",
                    "expiration_date",
                ]
            },
        ),
        (
            "Assignment",
            {
                "fields": [
                    "assignment",
                ]
            },
        ),
        (
            "Details",
            {
                "fields": [
                    "country_of_origin",
                    "manufactured_by",
                    "storage_conditions",
                ]
            },
        ),
        (
            "Reference",
            {
                "fields": [
                    "reference",
                    "comment",
                ]
            },
        ),
        audit_fieldset_tuple,
    )

    list_filter: tuple[str, ...] = (
        "lot_no",
        "expiration_date",
        "product",
        "assignment",
        "reference",
        "created",
        "modified",
    )

    list_display: tuple[str, ...] = (
        "lot_no",
        "expiration_date",
        "product",
        "assignment",
        "reference",
        "created",
        "modified",
    )
    radio_fields: tuple[str, ...] = {"assignment": admin.VERTICAL}

    search_fields: tuple[str, ...] = ("lot_no",)

    ordering: tuple[str, ...] = ("-expiration_date",)

    def get_readonly_fields(self, request, obj=None):
        # if obj:
        #     return self.readonly_fields + ("lot_no", "product", "assignment")
        return self.readonly_fields

    def get_fieldsets(self, request, obj=None):
        if obj and user_is_blinded_from_request(request):
            return (
                (
                    None,
                    {
                        "fields": [
                            "lot_no",
                            "expiration_date",
                            "product",
                        ]
                    },
                ),
                audit_fieldset_tuple,
            )
        return self.fieldsets

    def get_list_display(self, request):
        fields = super().get_list_display(request)
        fields = remove_fields_for_blinded_users(request, fields)
        return fields

    def get_list_filter(self, request):
        fields = super().get_list_filter(request)
        fields = remove_fields_for_blinded_users(request, fields)
        return fields

    def get_search_fields(self, request):
        fields = super().get_search_fields(request)
        fields = remove_fields_for_blinded_users(request, fields)
        return fields



# Generated by Django 2.2.3 on 2019-09-17 19:12

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_data_manager", "0006_auto_20190917_2133")]

    operations = [
        migrations.AlterField(
            model_name="datadictionary",
            name="model",
            field=models.CharField(help_text="label_lower", max_length=250),
        ),
        migrations.AlterField(
            model_name="historicaldatadictionary",
            name="model",
            field=models.CharField(help_text="label_lower", max_length=250),
        ),
    ]



from edc_identifier.simple_identifier import SimpleUniqueIdentifier


class ManifestIdentifier(SimpleUniqueIdentifier):
    random_string_length: int = 9
    identifier_type: str = "manifest_identifier"
    template: str = "M{device_id}{random_string}"



from ..df_handlers import CrfDfHandler
from .tables_exporter import TablesExporter


class CsvExporterNoTables(Exception):
    pass


class CsvCrfTablesExporterError(Exception):
    pass


class CsvCrfTablesExporter(TablesExporter):
    """A class to export CRF tables for this app_label.

    CRF tables include a FK to the visit model.
    """

    visit_column = None  # the visit column name used to select "CRF" tables
    df_handler_cls = CrfDfHandler

    def __init__(self, visit_column=None, with_columns=None, **kwargs):
        if visit_column:
            self.visit_column = visit_column
        if not self.visit_column:
            raise CsvCrfTablesExporterError("Visit column not specified. Got None.")
        with_columns = with_columns or []
        with_columns.extend([self.visit_column])
        super().__init__(with_columns=with_columns, **kwargs)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(app_label='{self.app_label}',"
            f"visit_column='{self.visit_column}')"
        )



from copy import copy

import pandas as pd

from ..constants import SYSTEM_COLUMNS
from ..mappings import yes_no
from .table import Table


class Aliquot(Table):
    default_columns = None
    requisition_columns = [
        "requisition_identifier",
        "panel_name",
        "requisition_datetime",
        "drawn_datetime",
        "gender",
        "dob",
        "study_site",
        "study_site_name",
        "visit_code",
        "visit_datetime",
        "visit_date",
    ]

    def prepare(self, df_requisition=None, requisition_columns=None, **kwargs):
        self.requisition_columns = requisition_columns or self.requisition_columns
        self.dataframe["aliquot_datetime"] = self.helper.to_local_datetime(
            self.dataframe["aliquot_datetime"]
        )
        self.dataframe["aliquot_date"] = self.dataframe["aliquot_datetime"].dt.normalize()

        # drop sys and other unwanted columns
        base_cols = copy(SYSTEM_COLUMNS)
        base_cols.append("id")
        base_cols.pop(base_cols.index("hostname_created"))
        base_cols = base_cols + ["slug", "comment", "shipped"]
        self.dataframe = self.dataframe.drop(base_cols, axis=1)

        # add column to indicate missing requisitions, remap columns
        self.dataframe["missing_requisition"] = self.missing_requisition(df_requisition)
        self.dataframe["is_primary"] = self.dataframe["is_primary"].map(
            yes_no, na_action="ignore"
        )
        self.dataframe["medium"] = self.dataframe["medium"].str.lower()
        self.dataframe["parent_identifier"] = self.dataframe.apply(
            lambda row: self.fix_parent_identifier(row["parent_identifier"]), axis=1
        )
        self.dataframe = pd.merge(
            self.dataframe,
            df_requisition[self.requisition_columns],
            on="requisition_identifier",
            how="left",
        )
        self.dataframe = self.dataframe[self.reordered_columns]

    @property
    def reordered_columns(self):
        columns = [
            col
            for col in list(self.dataframe.columns)
            if col
            not in [
                "hostname_created",
                "subject_identifier",
                "gender",
                "dob",
                "visit_code",
                "visit_datetime",
                "visit_date",
            ]
        ]
        columns.insert(0, "visit_code")
        columns.insert(0, "visit_date")
        columns.insert(0, "visit_datetime")
        columns.insert(0, "dob")
        columns.insert(0, "gender")
        columns.insert(0, "subject_identifier")
        columns.append("hostname_created")
        return columns

    def fix_parent_identifier(self, value):
        if value:
            value = value.replace("(", "").replace(")", "").replace("'", "").replace(",", "")
        return value

    def missing_requisition(self, df_requisition=None):
        """Returns a Series with True for rows missing a requisition."""
        missing = pd.merge(
            self.dataframe,
            df_requisition[self.requisition_columns],
            on="requisition_identifier",
            how="left",
        )
        return missing["panel_name"].isnull()



from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models


class SystolicPressureField(models.IntegerField):
    description = "Systolic pressure"

    def __init__(self, *args, verbose_name=None, **kwargs):
        kwargs["verbose_name"] = verbose_name or "Blood pressure: systolic"
        kwargs["validators"] = [MinValueValidator(50), MaxValueValidator(300)]
        kwargs["help_text"] = "in mm. format SYS, e.g. 120"
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["verbose_name"]
        del kwargs["validators"]
        del kwargs["help_text"]
        return name, path, args, kwargs


class DiastolicPressureField(models.IntegerField):
    description = "Diastolic pressure"

    def __init__(self, *args, verbose_name=None, **kwargs):
        kwargs["verbose_name"] = verbose_name or "Blood pressure: diastolic"
        kwargs["validators"] = [MinValueValidator(20), MaxValueValidator(225)]
        kwargs["help_text"] = "in Hg. format DIA, e.g. 80"
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["verbose_name"]
        del kwargs["validators"]
        del kwargs["help_text"]
        return name, path, args, kwargs



import arrow
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.utils import timezone

from edc_constants.constants import NO, OPEN, RESOLVED
from edc_lab.utils import get_requisition_model
from edc_metadata.constants import KEYED, REQUIRED
from edc_metadata.models import CrfMetadata, RequisitionMetadata
from edc_visit_tracking.utils import (
    get_related_visit_model_cls,
    get_subject_visit_missed_model_cls,
)

from ..constants import AUTO_RESOLVED
from ..models import DataQuery


class QueryRuleHandlerError(Exception):
    pass


class RequisitionNotKeyed(Exception):  # noqa: N818
    pass


class SpecimenNotDrawn(Exception):  # noqa: N818
    pass


class CrfInspectionFailed(Exception):  # noqa: N818
    pass


class DoNothingHandler:
    name = "do_nothing"
    display_name = "Do Nothing"

    def __init__(self, **kwargs):
        pass

    def run(self):
        pass


class QueryRuleHandler:
    """Called by the RuleRunner.

    Handles a single rule instance by either creating a
    new data query, resolving an existing data query,
    or doing nothing.
    """

    name = "default"
    display_name = "Default"
    model_name = None

    def __init__(
        self,
        query_rule_obj=None,
        registered_subject=None,
        visit_schedule_obj=None,
        visit_code_sequence=None,
        now=None,
    ):
        self._field_values = {}
        self._model_obj = None
        self._related_visit = None
        self._recipients = None
        self._requisition_obj = None
        self.created_counter = 0
        self.data_dictionaries = query_rule_obj.data_dictionaries.all()
        self.data_query = None
        self.model_cls = query_rule_obj.model_cls
        self.now = now
        self.query_rule_obj = query_rule_obj
        self.recipients = query_rule_obj.recipients.all()
        self.registered_subject = registered_subject
        self.resolved_counter = 0
        self.visit_schedule_obj = visit_schedule_obj
        self.visit_code_sequence = visit_code_sequence or 0

        if self.model_name and self.model_cls._meta.label_lower != self.model_name:
            raise QueryRuleHandlerError(
                f"Invalid model class for rule runner. Expected {self.model_name}. "
                f"Got {self.model_cls._meta.label_lower}"
            )

    def run(self):
        site_ids = [o.id for o in self.query_rule_obj.sites.all()]
        if not site_ids or settings.SITE_ID in site_ids:
            if self.resolved:
                self.data_query = self.get_or_create_data_query(get_only=True)
                if self.data_query:
                    self.resolve_existing_data_query()
            else:
                self.data_query = self.get_or_create_data_query()
                if self.data_query.site_resolved:
                    self.reopen_existing_data_query()
            self.resolve_if_missed_visit()

    @property
    def resolved(self):
        """Returns True if related_visit is NULL, or the query is not
        yet due, or the combination of field values is correct.

        To customize, try overriding `inspect_model`.
        """
        resolved = True
        if self.related_visit:
            try:
                self.inspect_requisition()
            except SpecimenNotDrawn:
                pass
            except RequisitionNotKeyed:
                resolved = not self.requisition_is_required
            else:
                if not self.model_obj:
                    resolved = False
                elif self.model_value_is_due:
                    if not self.crf_is_required:
                        resolved = False
                    else:
                        try:
                            self.inspect_model()
                        except CrfInspectionFailed:
                            resolved = False
        return resolved

    def resolve_if_missed_visit(self):
        if self.data_query:
            try:
                missed_visit_obj = get_subject_visit_missed_model_cls().objects.get(
                    subject_visit=self.related_visit
                )
            except ObjectDoesNotExist:
                pass
            else:
                self.resolve_existing_data_query(missed_visit_obj=missed_visit_obj)

    def inspect_requisition(self):
        """Raises a RequisitionNotKeyed or SpecimenNotDrawn exception
        if an expected Requisition is not keyed OR if a keyed
        requisition says specimen is not drawn.

        See `resolved`.
        """
        if self.query_rule_obj.requisition_panel:
            if not self.requisition_obj:
                raise RequisitionNotKeyed()
            if self.requisition_obj and self.requisition_obj.is_drawn == NO:
                raise SpecimenNotDrawn()

    def inspect_model(self):
        """Raises an CrfInspectionFailed exception if the combination
        of field values is incorrect.

        (Assumes a NULL field is missing, one with value is not)

        See `resolved`.

        May be overridden.
        """
        for field_name, field_value in self.field_values.items():
            if not field_value:
                raise CrfInspectionFailed(field_name)

    @property
    def requisition_is_required(self):
        """Returns True if a metadata instance of either REQUIRED
        or KEYED exists.
        """
        opts = dict(
            subject_identifier=self.registered_subject.subject_identifier,
            model=get_requisition_model()._meta.label_lower,
            panel_name=self.query_rule_obj.requisition_panel.name,
            visit_schedule_name=self.related_visit.visit_schedule_name,
            schedule_name=self.related_visit.schedule_name,
            visit_code=self.related_visit.visit_code,
            visit_code_sequence=self.related_visit.visit_code_sequence,
            entry_status__in=[REQUIRED, KEYED],
        )
        return RequisitionMetadata.objects.filter(**opts).exists()

    @property
    def crf_is_required(self):
        """Returns True if a metadata instance of either REQUIRED
        or KEYED exists.
        """
        return CrfMetadata.objects.filter(
            subject_identifier=self.registered_subject.subject_identifier,
            model=self.query_rule_obj.model_cls._meta.label_lower,
            visit_schedule_name=self.related_visit.visit_schedule_name,
            schedule_name=self.related_visit.schedule_name,
            visit_code=self.related_visit.visit_code,
            visit_code_sequence=self.related_visit.visit_code_sequence,
            entry_status__in=[REQUIRED, KEYED],
        ).exists()

    @property
    def model_value_is_due(self):
        """Returns True if value is expected relative to the
        timepoint report datetime.

        Returns True if the model is None.

        See fields `timing` and `timing_units`.
        """
        is_due = False
        if self.related_visit:
            if not self.model_obj:
                is_due = True
            else:
                now = arrow.get(self.now)
                start = arrow.get(self.related_visit.report_datetime)
                end = arrow.get(self.related_visit.report_datetime).shift(
                    **{self.query_rule_obj.timing_units: self.query_rule_obj.timing}
                )
                is_due = now >= start and not now.is_between(start, end, "[]")
        return is_due

    @property
    def field_values(self):
        if not self._field_values:
            for data_dictionary in self.data_dictionaries:
                if data_dictionary.field_name:
                    self._field_values.update(
                        {
                            data_dictionary.field_name: getattr(
                                self.model_obj, data_dictionary.field_name, None
                            )
                        }
                    )
        return self._field_values

    def get_field_value(self, field_name):
        """Safely get a model instance value for this query."""
        if not self.query_rule_obj.data_dictionaries.filter(field_name=field_name).exists():
            field_names = [
                f"{dd.field_name} ({dd.number})"
                for dd in self.query_rule_obj.data_dictionaries.all()
            ]
            raise QueryRuleHandlerError(
                f"Invalid field specified for query. Expected one of {field_names}. "
                f"Got {field_name}."
            )
        return getattr(self.model_obj, field_name)

    @property
    def resolved_datetime(self):
        return getattr(
            self.model_obj,
            "modified",
            getattr(self.requisition_obj, "modified", timezone.now()),
        )

    def resolve_existing_data_query(self, missed_visit_obj=None):
        """Resolves a data query model instance."""
        if self.data_query:
            site_response_text = (self.data_query.site_response_text or "").replace(
                AUTO_RESOLVED, ""
            )
            self.data_query.missed_visit = missed_visit_obj is not None
            self.data_query.site_response_text = f"{site_response_text} [auto-resolved]"
            self.data_query.auto_resolved = True
            self.data_query.site_resolved_datetime = self.resolved_datetime
            self.data_query.site_response_status = RESOLVED
            self.data_query.resolved_datetime = self.resolved_datetime
            self.data_query.dm_user = self.query_rule_obj.sender
            self.data_query.status = RESOLVED
            self.data_query.save()
            self.data_query.refresh_from_db()
            self.resolved_counter = 1

    def reopen_existing_data_query(self):
        """Re-opens a data query model instance unless locked."""
        if self.data_query and not self.data_query.locked:
            self.data_query.site_response_text.replace(AUTO_RESOLVED, "")
            self.data_query.site_resolved_datetime = None
            self.data_query.site_response_status = OPEN
            self.data_query.resolved_datetime = None
            self.data_query.missed_visit = False
            self.data_query.auto_resolved = False
            self.data_query.status = OPEN
            self.data_query.dm_user = None
            self.data_query.save()
            self.data_query.refresh_from_db()

    @property
    def related_visit(self):
        """Returns a visit model instance or None for this
        subject_identifier and visit schedule timepoint.
        """
        if not self._related_visit:
            try:
                self._related_visit = get_related_visit_model_cls().objects.get(
                    appointment__visit_schedule_name=(
                        self.visit_schedule_obj.visit_schedule_name
                    ),
                    appointment__schedule_name=self.visit_schedule_obj.schedule_name,
                    appointment__visit_code=self.visit_schedule_obj.visit_code,
                    appointment__visit_code_sequence=self.visit_code_sequence,
                    appointment__subject_identifier=(
                        self.registered_subject.subject_identifier
                    ),
                )
            except ObjectDoesNotExist:
                self._related_visit = None
        return self._related_visit

    @property
    def model_obj(self):
        """Returns a CRF model instance or None for this
        subject_identifier and visit schedule timepoint.
        """
        if not self._model_obj:
            try:
                self._model_obj = self.model_cls.objects.get(
                    **{f"{self.model_cls.related_visit_model_attr()}": self.related_visit}
                )
            except ObjectDoesNotExist:
                pass
            except AttributeError as e:
                if "related_visit_model_attr" in str(e):
                    self._model_obj = self.model_cls.objects.get(
                        subject_identifier=self.registered_subject.subject_identifier
                    )
        return self._model_obj

    @property
    def requisition_obj(self):
        """Returns a requisition model instance or None."""
        if not self._requisition_obj:
            try:
                opts = {
                    f"{self.model_cls.related_visit_model_attr()}": self.related_visit,
                    "panel": self.query_rule_obj.requisition_panel,
                }
                self._requisition_obj = get_requisition_model().objects.get(**opts)
            except ObjectDoesNotExist:
                pass
        return self._requisition_obj

    def get_or_create_data_query(self, get_only=None):
        """Returns a data_query.

        If one does not exist it will be created.
        """
        try:
            data_query = DataQuery.objects.get(
                subject_identifier=self.registered_subject.subject_identifier,
                rule_generated=True,
                rule_reference=self.query_rule_obj.reference,
                registered_subject=self.registered_subject,
                visit_schedule=self.visit_schedule_obj,
                site=self.registered_subject.site,
            )
        except ObjectDoesNotExist:
            data_query = None
            if not get_only:
                data_query = DataQuery(
                    query_priority=self.query_rule_obj.query_priority,
                    query_text=self.query_rule_obj.query_text,
                    registered_subject=self.registered_subject,
                    requisition_panel=self.query_rule_obj.requisition_panel,
                    rule_generated=True,
                    rule_reference=str(self.query_rule_obj.reference),
                    sender=self.query_rule_obj.sender,
                    site=self.registered_subject.site,
                    subject_identifier=self.registered_subject.subject_identifier,
                    title=self.query_rule_obj.title,
                    visit_schedule=self.visit_schedule_obj,
                )
                data_query.save()
                for data_dictionary in self.data_dictionaries:
                    data_query.data_dictionaries.add(data_dictionary)
                for recipient in self.recipients:
                    data_query.recipients.add(recipient)
                self.created_counter = 1
        return data_query



# Generated by Django 5.1 on 2024-09-09 14:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_action_item", "0037_remove_actionitem_reference_model_and_more"),
        (
            "edc_pharmacy",
            "0022_alter_historicalproduct_site_alter_historicalrx_site_and_more",
        ),
        ("sites", "0002_alter_domain_unique"),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name="rx",
            name="edc_pharmac_modifie_986021_idx",
        ),
        migrations.RemoveIndex(
            model_name="rx",
            name="edc_pharmac_user_mo_b9ac53_idx",
        ),
        migrations.AddIndex(
            model_name="rx",
            index=models.Index(fields=["rando_sid"], name="edc_pharmac_rando_s_0bcc84_idx"),
        ),
    ]



from django import forms

from edc_action_item.forms.action_item_form_mixin import ActionItemFormMixin

from ..models import DataDictionary


class DataDictionaryForm(ActionItemFormMixin, forms.ModelForm):

    class Meta:
        model = DataDictionary
        fields = "__all__"



from django.conf import settings

if (
    "edc_navbar" in settings.INSTALLED_APPS
    or "edc_navbar.apps.AppConfig" in settings.INSTALLED_APPS
):
    administration = ["edc_navbar.nav_administration"]
else:
    administration = []



from __future__ import annotations

import uuid

from .form_runner_by_scr_id import BaseFormRunnerBySrcId, FormRunnerBySrcId
from .site import site_form_runners

__all__ = ["get_form_runner_by_src_id"]


def get_form_runner_by_src_id(
    src_id: uuid.UUID | None = None,
    model_name: str | None = None,
    verbose: bool | None = None,
) -> BaseFormRunnerBySrcId | FormRunnerBySrcId:
    if form_runner_cls := site_form_runners.registry.get(model_name):
        return type("CustomFormRunnerBySrcId", (BaseFormRunnerBySrcId, form_runner_cls), {})(
            src_id=src_id, model_name=model_name
        )
    return FormRunnerBySrcId(src_id=src_id, model_name=model_name, verbose=verbose)



# Generated by Django 4.2.10 on 2024-03-07 15:11

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        (
            "edc_data_manager",
            "0037_alter_dataquery_site_alter_historicaldataquery_site_and_more",
        ),
    ]

    operations = [
        migrations.RenameField(
            model_name="datadictionary",
            old_name="default",
            new_name="default_value",
        ),
        migrations.RenameField(
            model_name="historicaldatadictionary",
            old_name="default",
            new_name="default_value",
        ),
    ]



# from importlib.metadata import version
#
# __version__ = version("edc_device")
#
# from .constants import CENTRAL_SERVER, CLIENT, MIDDLEMAN, NODE_SERVER
# from .device_permission import (
#     DeviceAddPermission,
#     DeviceChangePermission,
#     DevicePermissionAddError,
#     DevicePermissionChangeError,
#     DevicePermissions,
# )
#
# device_permissions = DevicePermissions()



from django.apps import AppConfig as DjangoApponfig
from django.core.management import color_style

style = color_style()


class AppConfig(DjangoApponfig):
    name = "edc_export"
    verbose_name = "Edc Export"
    include_in_administration_section = True
    default_auto_field = "django.db.models.BigAutoField"



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.http.response import HttpResponseRedirect
from django.urls.base import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.views.generic.base import TemplateView

from edc_dashboard.view_mixins import EdcViewMixin

from ..archive_exporter import (
    ArchiveExporter,
    ArchiveExporterEmailError,
    ArchiveExporterNothingExported,
)
from ..exportables import Exportables
from ..files_emailer import FilesEmailerError
from ..model_options import ModelOptions
from ..models import DataRequest, DataRequestHistory

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


class NothingToExport(Exception):
    pass


class ExportModelsViewError(Exception):
    pass


class ExportSelectedModelsView(EdcViewMixin, TemplateView):
    post_action_url = "edc_export:export_models_url"
    template_name = "edc_export/export_models.html"

    def get_context_data(self, **kwargs) -> dict:
        if self.request.session.get("selected_models"):
            kwargs.update(
                selected_models=[
                    ModelOptions(**dct) for dct in self.request.session["selected_models"]
                ]
            )
        return super().get_context_data(**kwargs)

    def post(self, request: WSGIRequest, *args, **kwargs) -> HttpResponseRedirect:
        if not self.check_user(request):
            pass
        elif not request.user.email:
            user_url = reverse("admin:auth_user_change", args=(request.user.id,))
            messages.error(
                request,
                format_html(
                    "Your account does not include an email address. "
                    'Please update your <a href="{user_url}">user account</a> '
                    "and try again.",
                    user_url=mark_safe(user_url),  # nosec B308 B703
                ),
            )
        else:
            try:
                self.export_models(request=request, email_to_user=True)
            except NothingToExport:
                messages.warning(
                    request,
                    "Nothing to do. Select one or more models and try again.",
                )
            except FilesEmailerError as e:
                messages.error(request, f"Failed to send the data you requested. Got '{e}'")
        url = reverse(self.post_action_url, kwargs=self.kwargs)
        return HttpResponseRedirect(url)

    @staticmethod
    def check_export_permissions(selected_models) -> list[ModelOptions]:
        return selected_models

    def export_models(self, request: WSGIRequest = None, email_to_user=None):
        selected_models = self.check_export_permissions(
            self.get_selected_models_from_session(request)
            or self.get_selected_models_from_post(request)
        )
        selected_models = [x.label_lower for x in selected_models]
        email_to_user = False if settings.DEBUG else email_to_user
        archive = True if settings.DEBUG else False
        try:
            exporter = ArchiveExporter(
                models=selected_models,
                user=request.user,
                email_to_user=email_to_user,
                archive=archive,
            )
        except (ArchiveExporterEmailError, ConnectionRefusedError) as e:
            messages.error(request, f"Failed to send files by email. Got '{e}'")
        except ArchiveExporterNothingExported:
            messages.info(request, "Nothing to export.")
        else:
            if email_to_user:
                msg = (
                    f"Your data request has been sent to {request.user.email}. "
                    "Please check your email."
                )
            else:
                msg = f"Your data request has been saved to {exporter.archive_filename}. "

            messages.success(request, msg)
            summary = [str(x) for x in exporter.exported]
            summary.sort()
            data_request = DataRequest.objects.create(
                name=f'Data request {datetime.now().strftime("%Y%m%d%H%M")}',
                models="\n".join(selected_models),
                user_created=request.user.username,
                site=request.site,
            )
            DataRequestHistory.objects.create(
                data_request=data_request,
                exported_datetime=exporter.exported_datetime,
                summary="\n".join(summary),
                user_created=request.user.username,
                user_modified=request.user.username,
                archive_filename=exporter.archive_filename,
                emailed_to=exporter.emailed_to,
                emailed_datetime=exporter.emailed_datetime,
                site=request.site,
            )

    @staticmethod
    def get_selected_models_from_post(request: WSGIRequest) -> list[ModelOptions]:
        """Returns a list of selected models from the POST
        as ModelOptions.
        """
        exportables = Exportables(request=request, user=request.user)
        selected_models = []
        for exportable in exportables:
            selected_models.extend(request.POST.getlist(f"chk_{exportable}_models") or [])
            selected_models.extend(
                request.POST.getlist(f"chk_{exportable}_historical_models") or []
            )
            selected_models.extend(request.POST.getlist(f"chk_{exportable}_list_models") or [])
            selected_models.extend(request.POST.getlist(f"chk_{exportable}_inlines") or [])
        if selected_models:
            selected_models = [ModelOptions(model=m) for m in selected_models if m]
        return selected_models

    @staticmethod
    def get_selected_models_from_session(request: WSGIRequest) -> list[ModelOptions]:
        """Returns a list of selected models from the session object
        as ModelOptions.
        """
        try:
            selected_models = request.session.pop("selected_models")
        except KeyError:
            selected_models = []
        else:
            selected_models = [ModelOptions(**dct) for dct in selected_models]
        return selected_models

    def check_user(self, request) -> bool:
        try:
            valid_user = User.objects.filter(username=self.request.user).exists()
        except ObjectDoesNotExist:
            messages.error(request, "Invalid user.")
            valid_user = False
        return valid_user



from __future__ import annotations

from edc_visit_tracking.utils import get_related_visit_model


class RuleGroupMetaError(Exception):
    pass


class RuleGroupMetaOptions:
    """Class to prepare the "meta" instance with the Meta class
    attributes.

    Adds default options if they were not declared on Meta class.

    """

    def __init__(self, group_name: str, attrs: dict) -> None:
        meta = attrs.pop("Meta", None)
        # assert metaclass was declared on the rule group
        if not meta:
            raise AttributeError(f"Missing Meta class. See {group_name}")
        # add default options if they do not exist
        for attr in self.default_meta_options:
            try:
                getattr(meta, attr)
            except AttributeError:
                setattr(meta, attr, None)
        # populate options dictionary
        self.options = {k: getattr(meta, k) for k in meta.__dict__ if not k.startswith("_")}
        # raise on any unknown attributes declared on the Meta class
        for meta_attr in self.options:
            if meta_attr not in [
                k for k in self.default_meta_options if not k.startswith("_")
            ]:
                raise RuleGroupMetaError(
                    f"Invalid _meta attr. Got '{meta_attr}'. See {group_name}."
                )
        # default app_label to current module if not declared
        module_name = attrs.get("__module__").split(".")[0]
        self.app_label = self.options.get("app_label", module_name)
        # source model
        self.source_model = self.options.get("source_model")
        if self.source_model:
            if len(self.source_model.split(".")) != 2:
                self.source_model = f"{self.app_label}.{self.source_model}"
            self.options.update(source_model=self.source_model)
        # related visit model
        self.related_visit_model = self.options.get("related_visit_model")
        if self.related_visit_model:
            if len(self.related_visit_model.split(".")) != 2:
                raise RuleGroupMetaError(
                    "Invalid _meta attr. Expected _meta.related_visit_model to be in "
                    f"label_lower format. Got '{self.related_visit_model}'. See {group_name}."
                )

            self.options.update(related_visit_model=self.related_visit_model)
        else:
            self.related_visit_model = get_related_visit_model()
        self.predicates = self.options.get("predicates", None)

    @property
    def default_meta_options(self) -> list[str]:
        return ["app_label", "source_model", "related_visit_model", "predicates"]



import re

import pandas as pd


def undash(value, exclude_pattern=None):
    if pd.notnull(value):
        try:
            if exclude_pattern:
                if not re.match(exclude_pattern, value):
                    value = value.replace("-", "")
            else:
                value = value.replace("-", "")
        except AttributeError as e:
            raise AttributeError(f"{e!s}. Got '{value}'")
    return value



from django.core.validators import RegexValidator
from django.utils.translation import gettext_lazy as _

from ..utils import dh_pattern, ymd_pattern

# expect 1h20m, 11h5m, etc
dh_validator = RegexValidator(
    dh_pattern,
    message=_(
        "Invalid format. Expected combinations of days and hours (dh): "
        "Something like 3d2h, 7d, 12h, etc. No spaces allowed."
    ),
)

hm_validator = RegexValidator(
    r"^([0-9]{1,3}h([0-5]?[0-9]m)?)$",
    message=_("Invalid format. Expected something like 1h20m, 11h5m, etc. No spaces allowed."),
)

ymd_validator = RegexValidator(
    ymd_pattern,
    message=_(
        "Invalid format. Expected combinations of years and months (ym): "
        "4y, 3y5m, 1y0m, 6m or days (d): 7d, 0d.  No spaces allowed."
    ),
)

hm_validator2 = RegexValidator(
    r"^([0-9]{1,3}:[0-5][0-9])$",
    message=_("Enter a valid time in hour:minutes format. No spaces allowed."),
)



from __future__ import annotations

import os
from textwrap import fill
from typing import TYPE_CHECKING

from bs4 import BeautifulSoup
from django.apps import apps as django_apps
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.utils.text import slugify
from reportlab.lib import colors
from reportlab.lib.pagesizes import A4
from reportlab.lib.units import cm
from reportlab.lib.utils import ImageReader
from reportlab.platypus import Paragraph, TableStyle
from reportlab.platypus.flowables import KeepTogether, Spacer
from reportlab.platypus.tables import Table

from edc_data_manager.get_longitudinal_value import (
    DataDictionaryError,
    get_longitudinal_value,
)
from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_randomization.auth_objects import RANDO_UNBLINDED
from edc_utils import formatted_age, get_static_file
from edc_utils.date import to_local

from .report import Report

if TYPE_CHECKING:
    from django.contrib.auth.models import User
    from django.core.handlers.wsgi import WSGIRequest

    from edc_crf.model_mixins import CrfModelMixin
    from edc_identifier.model_mixins import UniqueSubjectIdentifierModelMixin


class NotAllowed(Exception):
    pass


class CrfPdfReportError(Exception):
    pass


class CrfPdfReport(Report):
    model: str = None  # label_lower
    report_url: str = None
    changelist_url: str = None

    default_page = dict(
        rightMargin=1.5 * cm,
        leftMargin=1.5 * cm,
        topMargin=2.0 * cm,
        bottomMargin=1.5 * cm,
        pagesize=A4,
    )

    confidential = True
    draw_logo = True

    weight_model = None
    weight_field = "weight"

    open_label = True

    logo_data = {
        "app_label": "edc_pdf_reports",
        "filename": "clinicedc_logo.jpg",
        "first_page": (0.83 * cm, 0.83 * cm),
        "later_pages": (0.625 * cm, 0.625 * cm),
    }

    rando_user_group = None

    def __init__(
        self,
        model_obj: CrfModelMixin | UniqueSubjectIdentifierModelMixin = None,
        request: WSGIRequest | None = None,
        user: User | None = None,
        **extra_context,
    ):
        page: dict | None = extra_context.get("page")
        header_line: str | None = extra_context.get("header_line")
        filename: str | None = extra_context.get("filename")

        super().__init__(
            page=page, header_line=header_line, filename=filename, request=request
        )
        self._assignment = None
        self._logo = None
        self.model_obj = model_obj
        if not isinstance(self.model_obj, (self.get_model_cls(),)):
            raise CrfPdfReportError(
                f"Invalid model instance. Expected an instance of {self.get_model_cls()}. "
                f"Got {self.model_obj}."
            )
        if not self.changelist_url:
            raise CrfPdfReportError(f"Invalid changelist url. Got {self.changelist_url}.")

        self.user_model_cls = get_user_model()
        self.user = self.request.user if self.request else user
        self.subject_identifier = self.get_subject_identifier(**extra_context)
        self.bg_cmd = ("BACKGROUND", (0, 0), (0, -1), colors.lightgrey)

    def __repr__(self):
        return f"{self.__class__.__name__}(model_obj={self.model_obj})"

    def __str__(self):
        return self.model_obj

    def get_subject_identifier(self, **kwargs) -> str:
        try:
            subject_identifier = self.model_obj.related_visit.subject_identifier
        except AttributeError:
            subject_identifier = self.model_obj.subject_identifier
        return subject_identifier

    @property
    def report_filename(self) -> str:
        timestamp = to_local(self.model_obj.report_datetime).strftime("%Y%m%d")
        report_filename = (
            f"{slugify(self.model_obj._meta.verbose_name.lower())}-{self.subject_identifier}-"
            f"{timestamp}.pdf"
        )
        return report_filename

    @classmethod
    def get_generic_report_filename(cls) -> str:
        return f"{slugify(cls.get_verbose_name().lower())}s.pdf"

    @classmethod
    def get_model_cls(cls) -> type[CrfModelMixin | UniqueSubjectIdentifierModelMixin]:
        return django_apps.get_model(cls.model)

    @classmethod
    def get_verbose_name(cls) -> str:
        return cls.get_model_cls()._meta.verbose_name

    def get_report_story(self, **kwargs):
        story = []
        return story

    def on_first_page(self, canvas, doc):
        super().on_first_page(canvas, doc)
        width, height = A4
        if self.draw_logo and self.logo:
            canvas.drawImage(
                self.logo, 35, height - 50, *self.logo_data["first_page"], mask="auto"
            )
        else:
            canvas.setFontSize(10)
            canvas.drawString(48, height - 40, ResearchProtocolConfig().protocol_name)
        if self.confidential:
            canvas.setFontSize(10)
            canvas.drawString(48, height - 50, "CONFIDENTIAL")
            canvas.drawRightString(width - 35, height - 50, "CONFIDENTIAL")

        canvas.setFontSize(10)
        canvas.drawRightString(width - 35, height - 40, self.title)

    def on_later_pages(self, canvas, doc):
        super().on_later_pages(canvas, doc)
        width, height = A4
        if self.draw_logo and self.logo:
            canvas.drawImage(
                self.logo, 35, height - 40, *self.logo_data["later_pages"], mask="auto"
            )
        if self.confidential:
            canvas.setFontSize(10)
            canvas.drawRightString(width - 35, height - 45, "CONFIDENTIAL")
        if self.title:
            canvas.setFontSize(8)
            canvas.drawRightString(width - 35, height - 35, self.title)

    def draw_demographics(self, story, **kwargs):
        try:
            assignment = fill(self.assignment, width=80)
        except NotAllowed:
            assignment = "*****************"
        rows = [
            ["Subject:", self.subject_identifier],
            [
                "Gender/Age:",
                f"{self.registered_subject.get_gender_display()} {self.age}",
            ],
            ["Weight:", f"{self.weight_at_timepoint} kg"],
            [
                "Study site:",
                f"{self.registered_subject.site.id}: "
                f"{self.registered_subject.site.name.title()}",
            ],
            [
                "Randomization date:",
                self.registered_subject.randomization_datetime.strftime("%Y-%m-%d %H:%M"),
            ],
            ["Assignment:", assignment],
        ]

        t = Table(rows, (4 * cm, 14 * cm))
        self.set_table_style(t, bg_cmd=self.bg_cmd)
        t.hAlign = "LEFT"
        story.append(t)

    @staticmethod
    def set_table_style(t, bg_cmd=None):
        cmds = [
            ("INNERGRID", (0, 0), (-1, -1), 0.25, colors.black),
            ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
            ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
        ]
        if bg_cmd:
            cmds.append(bg_cmd)
        t.setStyle(TableStyle(cmds))
        t.hAlign = "LEFT"
        return t

    @staticmethod
    def history_change_message(obj):
        log_entry_model_cls = django_apps.get_model("admin.logentry")
        qs = log_entry_model_cls.objects.filter(
            action_time__gte=obj.modified, object_id=str(obj.id)
        ).order_by("action_time")
        log_entry = qs.first()
        try:
            soup = BeautifulSoup(log_entry.get_change_message(), features="html.parser")
            return soup.get_text()
        except AttributeError:
            return "--"

    def draw_narrative(self, story, title=None, text=None):
        t = Table([[title]], (18 * cm))
        self.set_table_style(t, bg_cmd=self.bg_cmd)
        p = Paragraph(text, self.styles["line_data_large"])
        p.hAlign = "LEFT"
        story.append(KeepTogether([t, Spacer(0.1 * cm, 0.5 * cm), p]))

    def draw_end_of_report(self, story):
        story.append(Paragraph("- End of report -", self.styles["line_label_center"]))

    @property
    def registered_subject(self):
        return django_apps.get_model("edc_registration.RegisteredSubject").objects.get(
            subject_identifier=self.subject_identifier
        )

    @property
    def logo(self):
        if not self._logo:
            path = get_static_file(self.logo_data["app_label"], self.logo_data["filename"])
            if os.path.isfile(path):
                self._logo = ImageReader(path)
        return self._logo

    @property
    def title(self):
        verbose_name = self.model_obj.verbose_name.upper()
        subject_identifier = self.model_obj.subject_identifier
        return f"{verbose_name} FOR {subject_identifier}"

    @property
    def weight_at_timepoint(self):
        """Returns weight in Kgs"""
        try:
            return get_longitudinal_value(
                subject_identifier=self.subject_identifier,
                reference_dt=self.model_obj.report_datetime,
                **self.get_weight_model_and_field(),
            )
        except DataDictionaryError:
            return ""

    def get_weight_model_and_field(self):
        return {"model": self.weight_model, "field": self.weight_field}

    @property
    def age(self):
        model_obj = self.model_obj
        return formatted_age(
            self.registered_subject.dob, reference_dt=model_obj.report_datetime
        )

    @property
    def unblinded(self):
        """Override to determine if assignment can be shown
        for this subject_identifier.

        Default: True
        """
        return True

    @property
    def assignment(self):
        """Returns the assignment from the Randomization List"""
        if not self._assignment:
            if (
                not self.unblinded
                or not self.user.groups.filter(name=RANDO_UNBLINDED).exists()
            ):
                raise NotAllowed(
                    "User does not have permissions to access randomization list. "
                    f"Got {self.user}"
                )
            randomization_list_model_cls = django_apps.get_model(
                self.registered_subject.randomization_list_model
            )
            self._assignment = randomization_list_model_cls.objects.get(
                subject_identifier=self.subject_identifier
            ).assignment_description
        return self._assignment

    def get_user(self, obj, field=None):
        field = field or "user_created"
        try:
            user = self.user_model_cls.objects.get(username=getattr(obj, field))
        except ObjectDoesNotExist:
            user_created = getattr(obj, field)
        else:
            user_created = f"{user.first_name} {user.last_name}"
        return user_created



from __future__ import annotations

from uuid import UUID

from celery import shared_task

from edc_utils.celery import celery_is_active, run_task_sync_or_async

from ..utils import process_repack_request


@shared_task
def process_repack_request_queryset(
    repack_request_pks: list[UUID], username: str = None
) -> None:
    if not celery_is_active():
        repack_request_pks = repack_request_pks[:1]
    for pk in repack_request_pks:
        run_task_sync_or_async(process_repack_request, repack_request_id=pk, username=username)


__all__ = ["process_repack_request_queryset"]



# Generated by Django 2.2.6 on 2019-10-24 01:26

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [("edc_randomization", "0002_historicalrandomizationlist")]

    operations = [
        migrations.AlterModelOptions(
            name="randomizationlist",
            options={
                "ordering": ("site_name", "sid"),
                "permissions": (("display_assignment", "Can display assignment"),),
            },
        )
    ]



from __future__ import annotations

from typing import TYPE_CHECKING
from urllib.parse import urlencode

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponseRedirect
from django.urls import NoReverseMatch, reverse

from edc_metadata.next_form_getter import NextFormGetter

from .base_model_admin_redirect_mixin import BaseModelAdminRedirectMixin

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


class ModelAdminNextUrlRedirectError(Exception):
    pass


class ModelAdminNextUrlError(Exception):
    pass


class ModelAdminNextUrlRedirectMixin(BaseModelAdminRedirectMixin):
    """Redirect on add, change, delete by reversing url_name
    in the querystring OR clicking Save Next.

    In your url &next=my_app:my_url_name,arg1,arg2&agr1=value1&arg2=
    value2&arg3=value3&arg4=value4...etc.
    """

    # this func is required if show_save_next=True
    # use edc_metadata.get_next_required_form
    next_form_getter_cls = NextFormGetter

    # need to override admin change_form template for these to wrrk
    show_save_next = False
    show_cancel = False

    next_querystring_attr = "next"

    def add_view(
        self,
        request: WSGIRequest,
        form_url: str = "",
        extra_context: dict | None = None,
    ):
        """Redirect before save on "cancel", otherwise return
        normal behavior.
        """
        if self.show_cancel and request.POST.get("_cancel"):
            redirect_url = self.get_next_redirect_url(request=request)
            return HttpResponseRedirect(redirect_url)
        extra_context = self.extra_context(extra_context)
        return super().add_view(request, form_url=form_url, extra_context=extra_context)

    def change_view(self, request: WSGIRequest, object_id, form_url="", extra_context=None):
        """Redirect before save on "cancel", otherwise return
        normal behavior.
        """
        if self.show_cancel and request.POST.get("_cancel"):
            redirect_url = self.get_next_redirect_url(request=request, object_id=object_id)
            return HttpResponseRedirect(redirect_url)
        extra_context = self.extra_context(extra_context)
        return super().change_view(
            request, object_id, form_url=form_url, extra_context=extra_context
        )

    def extra_context(self, extra_context=None):
        """Adds the booleans for the savenext and cancel buttons
        to the context.

        These are also referred to in the submit_line.html.
        """
        extra_context = extra_context or {}
        if self.show_save_next:
            extra_context.update(show_save_next=self.show_save_next)
        if self.show_cancel:
            extra_context.update(show_cancel=self.show_cancel)
        return extra_context

    def redirect_url(self, request, obj, post_url_continue=None) -> str | None:
        redirect_url = None
        if self.show_save_next and request.POST.get("_savenext"):
            redirect_url = self.get_savenext_redirect_url(request=request, obj=obj)
            if not redirect_url:
                redirect_url = self.get_next_redirect_url(request=request, obj=obj)
        elif (self.show_cancel and request.POST.get("_cancel")) or request.GET.get(
            self.next_querystring_attr
        ):
            redirect_url = self.get_next_redirect_url(request=request, obj=obj)
        if not redirect_url:
            redirect_url = super().redirect_url(
                request, obj, post_url_continue=post_url_continue
            )
        return redirect_url

    def get_next_redirect_url(self, request=None, **kwargs) -> str | None:
        """Returns a redirect url determined from the "next" attr
        in the querystring.
        """
        redirect_url = None
        if next_querystring := request.GET.get(self.next_querystring_attr):
            url_name = next_querystring.split(",")[0]
            options = self.get_next_options(request=request, **kwargs)
            try:
                redirect_url = reverse(url_name, kwargs=options)
            except NoReverseMatch as e:
                msg = f"{e}. Got url_name={url_name}, kwargs={options}."
                try:
                    redirect_url = reverse(url_name)  # retry without kwargs
                except NoReverseMatch:
                    # raise with first exception msg
                    raise ModelAdminNextUrlRedirectError(msg)
                else:
                    if "q" in options and "changelist" in url_name:
                        redirect_url = f"{redirect_url}?q={options['q']}"
        return redirect_url

    def get_savenext_redirect_url(self, request=None, obj=None) -> str | None:
        """Returns a redirect_url for the next form in
        the visit schedule.

        This method expects a CRF model with model mixins
        from edc_visit_tracking and edc_visit_schedule.

        Requires edc_metadata. Queries Metadata models.
        """
        panel_name = None
        redirect_url = self.get_next_redirect_url(request=request)
        getter = self.next_form_getter_cls(model_obj=obj)
        if getter.next_form:
            try:
                panel_name = getter.next_form.panel.name
            except AttributeError:
                panel_name = None
            next_model_cls = django_apps.get_model(getter.next_form.model)
            url_name = "_".join(next_model_cls._meta.label_lower.split("."))
            url_name = f"{self.admin_site.name}:{url_name}"
            lookup_opts = {obj.related_visit_model_attr(): obj.related_visit}
            if panel_name:
                lookup_opts.update(panel__name=panel_name)
            try:
                next_obj = next_model_cls.objects.get(**lookup_opts)
            except ObjectDoesNotExist:
                redirect_url = reverse(f"{url_name}_add")
            else:
                redirect_url = reverse(f"{url_name}_change", args=(next_obj.id,))
        next_querystring = request.GET.get(self.next_querystring_attr)
        querystring_opts = self.get_next_options(request=request)
        querystring_opts.update({obj.related_visit_model_attr(): str(obj.related_visit.id)})
        if panel_name:
            panel_model_cls = django_apps.get_model("edc_lab.panel")
            panel = panel_model_cls.objects.get(name=panel_name)
            querystring_opts.update(panel=str(panel.id))
        querystring = urlencode(querystring_opts)
        if redirect_url:
            return (
                f"{redirect_url}?{self.next_querystring_attr}={next_querystring}&{querystring}"
            )
        return None

    def get_next_options(self, request=None, **kwargs):
        """Returns the key/value pairs from the "next" querystring
        as a dictionary.
        """
        attrs = request.GET.get(self.next_querystring_attr).split(",")[1:]
        return {k: request.GET.get(k) for k in attrs if request.GET.get(k)}



from typing import Any

from django.views.generic.base import TemplateView

from edc_dashboard.view_mixins import EdcViewMixin
from edc_navbar import NavbarViewMixin

from ..admin_site import edc_consent_admin
from ..site_consents import site_consents


class HomeView(EdcViewMixin, NavbarViewMixin, TemplateView):
    template_name = "edc_consent/home.html"
    navbar_name = "edc_consent"
    navbar_selected_item = "consent"

    def __init__(self, *args, **kwargs):
        super(HomeView, self).__init__(*args, **kwargs)

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        kwargs.update(
            edc_consent_admin=edc_consent_admin,
            consent_definitions=site_consents.get_consent_definitions,
        )
        return super().get_context_data(**kwargs)



from typing import Any

from django.forms import ValidationError

from edc_constants.constants import OTHER

from .base_form_validator import NOT_REQUIRED_ERROR, REQUIRED_ERROR, BaseFormValidator


class OtherSpecifyFieldValidator(BaseFormValidator):
    """A modelform mixin that handles 'OTHER/Other specify'
    field pattern.
    """

    def validate_other_specify(
        self,
        field: str,
        other_specify_field: str | None = None,
        required_msg: str | None = None,
        not_required_msg: str | None = None,
        other_stored_value: Any | None = None,
        ref: str | None = None,
        fk_stored_field_name: str | None = None,
    ) -> bool:
        """Returns False or raises a ValidationError.

        Note: "stored" means value stored in the db table as
              opposed to the "display" value.
        """
        cleaned_data = self.cleaned_data
        other = other_stored_value or OTHER
        if fk_stored_field_name is None:
            fk_stored_field_name = self.default_fk_stored_field_name

        # assume field naming convention
        if not other_specify_field:
            other_specify_field = f"{field}_other"

        # perhaps this is a list field / fk
        field_value = getattr(
            cleaned_data.get(field), fk_stored_field_name, cleaned_data.get(field)
        )

        if field_value and field_value == other and not cleaned_data.get(other_specify_field):
            ref = "" if not ref else f" ref: {ref}"
            message = {other_specify_field: required_msg or f"This field is required.{ref}"}
            self._errors.update(message)
            self._error_codes.append(REQUIRED_ERROR)
            raise ValidationError(message, code=REQUIRED_ERROR)
        if (
            field_value and field_value != other and cleaned_data.get(other_specify_field)
        ) or (field_value is None and cleaned_data.get(other_specify_field)):
            ref = "" if not ref else f" ref: {ref}"
            message = {
                other_specify_field: not_required_msg or f"This field is not required.{ref}"
            }
            self._errors.update(message)
            self._error_codes.append(NOT_REQUIRED_ERROR)
            raise ValidationError(message, code=NOT_REQUIRED_ERROR)
        return False



def post_migrate_update_notifications(sender=None, **kwargs):
    from .site_notifications import site_notifications

    site_notifications.update_notification_list(verbose=True)
    site_notifications.create_mailing_lists(verbose=True)



from edc_constants.constants import GRADE3, GRADE4, NO, NOT_APPLICABLE

from .constants import ALREADY_REPORTED, PRESENT_AT_BASELINE

REPORTABLE = (
    (NOT_APPLICABLE, "Not applicable"),
    (GRADE3, "Yes, grade 3"),
    (GRADE4, "Yes, grade 4"),
    (NO, "Not reportable"),
    (ALREADY_REPORTED, "Already reported"),
    (PRESENT_AT_BASELINE, "Present at baseline"),
)



# Generated by Django 3.1.5 on 2021-02-03 04:06

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_action_item", "0028_auto_20210203_0706"),
        ("edc_data_manager", "0017_auto_20200729_2310"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dataquery",
            name="action_item",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_action_item.actionitem",
            ),
        ),
        migrations.AlterField(
            model_name="dataquery",
            name="parent_action_identifier",
            field=models.CharField(
                blank=True,
                help_text="action identifier that links to parent reference model instance.",
                max_length=30,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="dataquery",
            name="parent_action_item",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="edc_action_item.actionitem",
            ),
        ),
        migrations.AlterField(
            model_name="dataquery",
            name="related_action_identifier",
            field=models.CharField(
                blank=True,
                help_text="action identifier that links to related reference model instance.",
                max_length=30,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="dataquery",
            name="related_action_item",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="edc_action_item.actionitem",
            ),
        ),
        migrations.AlterField(
            model_name="historicaldataquery",
            name="parent_action_identifier",
            field=models.CharField(
                blank=True,
                help_text="action identifier that links to parent reference model instance.",
                max_length=30,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="historicaldataquery",
            name="related_action_identifier",
            field=models.CharField(
                blank=True,
                help_text="action identifier that links to related reference model instance.",
                max_length=30,
                null=True,
            ),
        ),
    ]



# Generated by Django 2.2.6 on 2019-10-26 19:31

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_adverse_event", "0002_auto_20190802_0059")]

    operations = [
        migrations.AlterModelOptions(
            name="aeclassification",
            options={"ordering": ["display_index", "display_name"]},
        ),
        migrations.AlterModelOptions(
            name="causeofdeath", options={"ordering": ["display_index", "display_name"]}
        ),
        migrations.AlterModelOptions(
            name="saereason", options={"ordering": ["display_index", "display_name"]}
        ),
        migrations.RemoveIndex(
            model_name="aeclassification", name="edc_adverse_id_8133d9_idx"
        ),
        migrations.RemoveIndex(model_name="causeofdeath", name="edc_adverse_id_42a796_idx"),
        migrations.RemoveIndex(model_name="saereason", name="edc_adverse_id_88072b_idx"),
        migrations.RenameField(
            model_name="aeclassification", old_name="name", new_name="display_name"
        ),
        migrations.RenameField(
            model_name="causeofdeath", old_name="name", new_name="display_name"
        ),
        migrations.RenameField(
            model_name="saereason", old_name="name", new_name="display_name"
        ),
        migrations.RenameField(
            model_name="aeclassification", old_name="short_name", new_name="name"
        ),
        migrations.RenameField(
            model_name="causeofdeath", old_name="short_name", new_name="name"
        ),
        migrations.RenameField(model_name="saereason", old_name="short_name", new_name="name"),
        migrations.AddIndex(
            model_name="aeclassification",
            index=models.Index(
                fields=["id", "display_name", "display_index"],
                name="edc_adverse_id_6431b6_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="causeofdeath",
            index=models.Index(
                fields=["id", "display_name", "display_index"],
                name="edc_adverse_id_ca88f1_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="saereason",
            index=models.Index(
                fields=["id", "display_name", "display_index"],
                name="edc_adverse_id_dc0e62_idx",
            ),
        ),
    ]



class PrescriptionApprovalValidatorError(Exception):
    pass


class ApprovePrescription:
    prescription_model = "edc_pharmacy.rx"

    def __init__(self, rx_model_obj=None):
        dispense_appointment = rx_model_obj.dispense_appointment
        previous_appointment = dispense_appointment.previous()
        if previous_appointment:
            if rx_model_obj.__class__.objects.filter(
                is_approved=False, dispense_appointment=previous_appointment
            ).exists():
                raise PrescriptionApprovalValidatorError(
                    f"Future prescriptions cannot be approved. Approve "
                    f"prescriptions for dispensing date "
                    f"{previous_appointment.appt_datetime}"
                )



from django.views.generic import TemplateView

from edc_navbar import NavbarViewMixin

from ..view_mixins import AdministrationViewMixin, EdcViewMixin


class AdministrationView(EdcViewMixin, NavbarViewMixin, AdministrationViewMixin, TemplateView):
    navbar_selected_item = "administration"
    navbar_name = "default"  # settings.APP_NAME



# Generated by Django 4.2.7 on 2023-12-03 00:22

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "edc_unblinding",
            "0010_alter_historicalunblindingrequest_device_created_and_more",
        ),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="unblindingrequest",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Unblinding Request",
                "verbose_name_plural": "Unblinding Requests",
            },
        ),
        migrations.AlterModelOptions(
            name="unblindingreview",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Unblinding Review",
                "verbose_name_plural": "Unblinding Reviews",
            },
        ),
        migrations.AddField(
            model_name="historicalunblindingrequest",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalunblindingrequest",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalunblindingreview",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalunblindingreview",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="unblindingrequest",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="unblindingrequest",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="unblindingreview",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="unblindingreview",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
    ]



from django.contrib import admin
from django_pylabels.actions import copy_label_specification, export_to_csv
from django_revision.modeladmin_mixin import ModelAdminRevisionMixin

from edc_model_admin.mixins import (
    ModelAdminFormAutoNumberMixin,
    ModelAdminFormInstructionsMixin,
    ModelAdminInstitutionMixin,
    ModelAdminNextUrlRedirectMixin,
    ModelAdminRedirectOnDeleteMixin,
    TemplatesModelAdminMixin,
)

from ..admin_site import edc_pylabels_admin
from ..models import LabelSpecification


@admin.register(LabelSpecification, site=edc_pylabels_admin)
class LabelSpecificationAdmin(
    TemplatesModelAdminMixin,
    ModelAdminNextUrlRedirectMixin,  # add
    ModelAdminFormInstructionsMixin,  # add
    ModelAdminFormAutoNumberMixin,
    ModelAdminRevisionMixin,  # add
    ModelAdminInstitutionMixin,  # add
    ModelAdminRedirectOnDeleteMixin,
    admin.ModelAdmin,
):
    actions = [copy_label_specification, export_to_csv]

    instructions = (
        "This model captures the dimensions, rows, columns, "
        "and spacing for label sheet stationery."
    )

    date_hierarchy = "created"

    list_display = (
        "name",
        "page_description",
        "layout_description",
        "label_description",
        "border",
    )

    readonly_fields = (
        "page_description",
        "layout_description",
        "label_description",
    )

    list_filter = ("created", "modified")



import pandas as pd
from django.conf import settings

from edc_constants.constants import NO, YES


class Helper:
    to_dataframe_cls = None

    def __init__(self, local_tz=None, date_format=None):
        self.date_format = date_format or "%Y-%m-%d %H:%M:%S.%f"
        self.local_tz = local_tz or settings.TIME_ZONE

    def get_crf_dataframe(self, model=None, cols=None):
        """Returns a CRF dataframe characterized by having an index set
        to 'subject_visit_id'.
        """
        df_crf = self.to_dataframe_cls(model=model, drop_sys_columns=True).dataframe
        if cols:
            cols.append("subject_visit_id")
            df_crf = df_crf[cols]
        df_crf = df_crf.set_index("subject_visit_id")
        return df_crf

    def to_local_datetime(self, s):
        """Returns a localized datetime series given tz-aware dates."""
        s = s.dt.tz_localize("UTC")
        s = s.dt.tz_convert(self.local_tz)
        return s

    def date_to_local_datetime(self, s):
        """Returns a localized datetime series given naive dates."""
        s = pd.to_datetime(s, format=self.date_format, exact=True, utc=True, box=True)
        s = s.dt.tz_localize(self.local_tz)
        return s

    def get_yesno_etc(self, value):
        """Returns a value as Yes/No/not_sure given
        1, 0 True, False, etc.
        """
        try:
            value = str(int(value))
        except ValueError:
            pass
        if value in ["1", "True"]:
            value = YES
        elif value in ["0", "False"]:
            value = NO
        elif value == "2":
            value = "not_sure"
        else:
            pass
        return value



from django.core.validators import MinValueValidator
from django.db import models
from django.utils import timezone

from edc_constants.choices import YES_NO
from edc_model import models as edc_models
from edc_protocol.validators import date_not_before_study_start

from .choices import LTFU_CHOICES


class LtfuModelMixin(models.Model):
    report_datetime = models.DateTimeField(
        verbose_name="Report Date and Time",
        default=timezone.now,
        validators=[date_not_before_study_start],
    )

    last_seen_datetime = models.DateField(
        verbose_name="Date participant last seen",
        validators=[date_not_before_study_start],
    )

    number_consecutive_missed_visits = models.IntegerField(
        verbose_name="Number of consecutive visits missed", null=True, blank=False
    )

    last_missed_visit_datetime = models.DateField(
        verbose_name="Date of last missed visit report submitted",
        null=True,
        blank=False,
    )

    ltfu_date = models.DateField(
        verbose_name="Date participant considered lost to follow up",
        default=timezone.now,
        null=True,
        blank=False,
    )

    phone = models.CharField(
        verbose_name="Was contact by phone attempted",
        max_length=15,
        choices=YES_NO,
        default="",
        blank=False,
    )
    phone_attempts = models.IntegerField(
        verbose_name=(
            "If YES, how many attempts were made to contact the participant by phone"
        ),
        validators=[MinValueValidator(0)],
        default=0,
    )

    home_visited = models.CharField(
        verbose_name="Was a home visit attempted", max_length=15, choices=YES_NO
    )

    home_visit_detail = models.TextField(
        verbose_name="If YES, provide any further details of the home visit",
        blank=True,
        default="",
    )

    ltfu_category = models.CharField(
        verbose_name="Category of loss to follow up",
        max_length=25,
        choices=LTFU_CHOICES,
    )

    ltfu_category_other = edc_models.OtherCharField()

    comment = models.TextField(
        verbose_name=(
            "If any, please give additional details of the "
            "circumstances that led to this decision."
        ),
        blank=True,
        default="",
    )

    class Meta:
        abstract = True
        verbose_name = "Loss to Follow Up"
        verbose_name_plural = "Loss to Follow Ups"



from django.db import models


class ExportData(models.Model):
    """Dummy model for permissions"""

    class Meta:
        permissions = [("display_export_admin_action", "Display export action")]


class ImportData(models.Model):
    """Dummy model for permissions"""

    class Meta:
        permissions = [("display_import_admin_action", "Display import action")]



from edc_auth.constants import EVERYONE
from edc_auth.site_auths import site_auths
from edc_auth.utils import remove_default_model_permissions_from_edc_permissions

from .auth_objects import custom_codename_tuples


def update_site_auths():
    site_auths.add_post_update_func(
        "edc_navbar", remove_default_model_permissions_from_edc_permissions
    )
    site_auths.add_custom_permissions_tuples(
        model="edc_navbar.edcpermissions", codename_tuples=custom_codename_tuples
    )

    site_auths.update_group(
        "edc_navbar.nav_administration",
        "edc_navbar.nav_home",
        "edc_navbar.nav_logout",
        "edc_navbar.nav_public",
        name=EVERYONE,
    )


update_site_auths()



import sys
from inspect import isfunction
from warnings import warn

from django.apps import apps as django_apps
from django.contrib.admin import sites
from django.core.exceptions import FieldDoesNotExist, ObjectDoesNotExist
from django.db.models.fields import NOT_PROVIDED
from django.db.utils import IntegrityError, OperationalError
from django.test.client import RequestFactory
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from edc_list_data.model_mixins import ListModelMixin

from .models import DataDictionary, DataManagerUser

WIDGET = 1
FIELD_NAME = 0


class DbField:
    def __init__(self, model, fld):
        try:
            db_field = model._meta.get_field(get_field_name(fld))
        except FieldDoesNotExist:
            pass
        else:
            self.decimal_places = getattr(db_field, "decimal_places", None)
            self.default = self.get_default(db_field)
            self.field_name = getattr(db_field, "name", None)
            self.field_type = None if db_field is None else db_field.get_internal_type()
            self.help_text = format_html(
                "{}",
                mark_safe(
                    getattr(db_field, "help_text", ""),
                ),  # nosec B703, B308
            )
            self.max_digits = getattr(db_field, "max_digits", None)
            self.max_length = getattr(db_field, "max_length", None)
            self.nullable = getattr(db_field, "null", False)

    @staticmethod
    def get_default(db_field):
        default = getattr(db_field, "default", None)
        if default == NOT_PROVIDED:
            default = None
        if isfunction(default):
            default = default.__name__
        return default


def get_field_name(fld):
    try:
        return fld[FIELD_NAME]
    except TypeError:
        return fld.name


def get_form_field_type(model, fld):
    field_type = None
    try:
        db_field = model._meta.get_field(fld[FIELD_NAME])
    except FieldDoesNotExist as e:
        warn(str(e))
    else:
        field_type = db_field.get_internal_type()
    return field_type


def get_form_label(fld):
    try:
        label = str(fld[WIDGET].original_label)
    except AttributeError:
        label = str(fld[WIDGET].label)
    return label


def create_or_update_data_dictionary(index, model, fld):
    data_dictionary_model_cls = django_apps.get_model("edc_data_manager.datadictionary")
    try:
        label = get_form_label(fld)
    except TypeError:
        label = fld.verbose_name
    db_field_data = DbField(model, fld).__dict__
    if db_field_data.get("field_name"):
        default_value = db_field_data.pop("default")
        options = dict(
            active=True,
            model=model._meta.label_lower,
            number=index + 1,
            prompt=format_html("{}", mark_safe(label or "")),  # nosec B703, B308
            default_value=default_value,
            **db_field_data,
        )
        try:
            obj = data_dictionary_model_cls.objects.get(
                field_name=get_field_name(fld), model=model._meta.label_lower
            )
        except ObjectDoesNotExist:
            try:
                data_dictionary_model_cls.objects.create(**options)
            except (OperationalError, IntegrityError) as e:
                warn(
                    f"Error when creating DataDictionary instance. " f"Model={model}. Got {e}"
                )
        else:
            for k, v in options.items():
                setattr(obj, k, v)
            obj.save()


def populate_data_dictionary(form=None, model=None):
    auto_number = True
    try:
        auto_number = form._meta.auto_number
    except AttributeError:
        auto_number = True
    finally:
        if auto_number:
            for index, fld in enumerate(form.base_fields.items()):
                create_or_update_data_dictionary(index, model, fld)


def populate_data_dictionary_from_sites(request=None):
    """Populates the DataDictionary model using ModelAdmin
    classes.

    Set ModelAdmin.exclude_from_data_dictionary = True to exclude a model.

    If a form class has no base_fields falls back on editable model class fields.
    """

    app_labels = [app_config.name for app_config in django_apps.get_app_configs()]
    if not request:
        rf = RequestFactory()
        request = rf.get("/")
        request.user = DataManagerUser(
            id=1, username="erikvw", is_superuser=True, is_active=True
        )
    DataDictionary.objects.update(active=False)
    for site in sites.all_sites.data:
        sys.stdout.write(f" * {site().name} ...\n")
        for model_admin in site()._registry.values():
            form = model_admin.get_form(request)
            model = model_admin.model
            if model._meta.app_label in app_labels and not issubclass(
                model, (ListModelMixin,)
            ):
                populate = getattr(model_admin, "populate_data_dictionary", True)
                if not populate:
                    sys.stdout.write(f"   - {model._meta.label_lower}. (skipping)\n")
                else:
                    sys.stdout.write(f"   + {model._meta.label_lower}...\n")
                    if not form.base_fields:
                        for index, fld in enumerate(model._meta.get_fields()):
                            if fld.editable:
                                create_or_update_data_dictionary(index, model, fld)
                    else:
                        for index, fld in enumerate(form.base_fields.items()):
                            create_or_update_data_dictionary(index, model, fld)
                    sys.stdout.write(f"   + {model._meta.label_lower}.   \n")



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django.utils import timezone

from edc_dashboard.view_mixins import EdcViewMixin
from edc_listboard.view_mixins import ListboardFilterViewMixin, SearchFormViewMixin
from edc_listboard.views import ListboardView as BaseListboardView
from edc_navbar import NavbarViewMixin

from ...constants import (
    AE_FOLLOWUP_ACTION,
    AE_TMG_ACTION,
    DEATH_REPORT_ACTION,
    DEATH_REPORT_TMG_ACTION,
)
from ...utils import get_adverse_event_app_label

if TYPE_CHECKING:
    from django.db.models import Q


class SummaryListboardView(
    NavbarViewMixin,
    EdcViewMixin,
    ListboardFilterViewMixin,
    SearchFormViewMixin,
    BaseListboardView,
):
    listboard_back_url = "tmg_home_url"

    ae_tmg_model = f"{get_adverse_event_app_label()}.aetmg"
    listboard_template = "tmg_summary_listboard_template"
    listboard_url = "tmg_summary_listboard_url"
    listboard_panel_style = "warning"
    listboard_model = "edc_action_item.actionitem"
    listboard_panel_title = "TMG: Events Summary"
    listboard_view_permission_codename = "edc_adverse_event.view_tmg_listboard"

    navbar_selected_item = "tmg_home"
    ordering = "-report_datetime"
    paginate_by = 25
    search_form_url = "tmg_summary_listboard_url"
    action_type_names = [
        AE_TMG_ACTION,
        DEATH_REPORT_TMG_ACTION,
        DEATH_REPORT_ACTION,
        AE_FOLLOWUP_ACTION,
    ]
    search_fields = [
        "subject_identifier",
        "action_identifier",
        "parent_action_item__action_identifier",
        "related_action_item__action_identifier",
        "user_created",
        "user_modified",
    ]

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        kwargs.update(AE_TMG_ACTION=AE_TMG_ACTION, utc_date=timezone.now().date())
        return super().get_context_data(**kwargs)

    def get_queryset_filter_options(self, request, *args, **kwargs) -> tuple[Q, dict]:
        q_object, options = super().get_queryset_filter_options(request, *args, **kwargs)
        options.update({"action_type__name__in": self.action_type_names})
        return q_object, options



from django.apps import apps as django_apps
from django.core.checks import CheckMessage, Warning

from .metadata_rules import site_metadata_rules


def check_for_metadata_rules(app_configs, **kwargs) -> list[CheckMessage]:
    errors = []
    if not site_metadata_rules.registry:
        errors.append(
            Warning(
                "No metadata rules were loaded by site_metadata_rules.autodiscover.",
                id="edc_metadata.W001",
            )
        )
    if not django_apps.get_app_config("edc_metadata").metadata_rules_enabled:
        errors.append(Warning("Metadata rules disabled!", id="edc_metadata.W002"))
    return errors



from decimal import Decimal

from django.contrib.admin import SimpleListFilter

__all__ = ["FbgListFilter", "OgttListFilter"]


class FbgListFilter(SimpleListFilter):
    title = "FBG"
    parameter_name = "fbg_value"

    def lookups(self, request, model_admin):
        return (
            ("fbg_normal", "below 7.0 mmol/L"),
            ("fbg_high", "above 7.0 mmol/L (incl)"),
        )

    def queryset(self, request, queryset):
        self.value()
        if self.value() == "fbg_normal":
            return queryset.filter(fbg_value__lt=Decimal("7.0"))
        if self.value() == "fbg_high":
            return queryset.filter(fbg_value__gte=Decimal("7.0"))
        return queryset


class OgttListFilter(SimpleListFilter):
    title = "OGTT"
    parameter_name = "ogtt_value"

    def lookups(self, request, model_admin):
        return (
            ("ogtt_normal", "below 11.1 mmol/L"),
            ("ogtt_high", "above 11.1 mmol/L (incl)"),
        )

    def queryset(self, request, queryset):
        if self.value() == "ogtt_normal":
            return queryset.filter(ogtt_value__lt=Decimal("11.1"))
        if self.value() == "ogtt_high":
            return queryset.filter(ogtt_value__gte=Decimal("11.1"))
        return queryset



from django.utils.html import format_html
from django.utils.safestring import mark_safe


class FormLabelModelAdminMixin:
    """A class that helps modify fieldsets for subject models

    * Model is expected to have a relation `subject_visit__appointment`.
    * Expects appointment to be in GET
    """

    custom_form_labels = []

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj=obj, **kwargs)
        return self.update_form_labels(request=request, obj=obj, form=form)

    def update_form_labels(self, request=None, obj=None, form=None):
        """Returns a form obj after modifying form labels
        referred to in custom_form_labels.
        """
        for form_label in self.custom_form_labels:
            if form_label.field in form.base_fields:
                label = form_label.get_form_label(
                    request=request, obj=obj, model=self.model, form=form
                )
                if label:
                    form.base_fields[form_label.field].label = format_html(
                        "{label}",
                        label=mark_safe(label),  # nosec B308, B703
                    )
        return form



from gettext import gettext as _

from django import forms

from ..calculators import CalculatorError, calculate_bmi


class WeightHeightBmiFormValidatorMixin:
    @staticmethod
    def validate_weight_height_with_bmi(weight_kg=None, height_cm=None, **kwargs):
        try:
            bmi = calculate_bmi(weight_kg=weight_kg, height_cm=height_cm, **kwargs)
        except CalculatorError as err:
            errmsg = _("Please check weight and height. {err}").format(err=err)
            raise forms.ValidationError(errmsg) from err
        return bmi



from copy import copy
from typing import Any

from django.urls import reverse

from edc_dashboard.url_names import url_names
from edc_lab.constants import SHIPPED

from .base_box_item_listboard_view import BaseBoxItemListboardView


class VerifyBoxListboardView(BaseBoxItemListboardView):
    action_name = "verify"
    form_action_url = "verify_box_item_form_action_url"
    listboard_template = "verify_box_listboard_template"
    listboard_url = "verify_box_listboard_url"
    navbar_selected_item = "pack"
    search_form_url = "verify_box_listboard_url"
    manage_box_listboard_url = "manage_box_listboard_url"
    verify_box_listboard_url = "verify_box_listboard_url"

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        kwargs.update(
            manage_box_listboard_url_reversed=self.manage_box_listboard_url_reversed,
            verify_box_listboard_url_reversed=self.verify_box_listboard_url_reversed,
            position=self.kwargs.get("position"),
            SHIPPED=SHIPPED,
        )
        return super().get_context_data(**kwargs)

    @property
    def verify_box_listboard_url_reversed(self):
        url_kwargs = copy(self.url_kwargs)
        url_kwargs["position"] = 1
        url_kwargs["action_name"] = "verify"
        return reverse(url_names.get(self.verify_box_listboard_url), kwargs=url_kwargs)

    @property
    def manage_box_listboard_url_reversed(self):
        url_kwargs = copy(self.url_kwargs)
        url_kwargs.pop("position")
        url_kwargs["action_name"] = "manage"
        return reverse(url_names.get(self.manage_box_listboard_url), kwargs=url_kwargs)

    @property
    def url_kwargs(self):
        return {
            "action_name": self.action_name,
            "box_identifier": self.box_identifier,
            "position": self.kwargs.get("position", "1"),
        }



import pandas as pd

from ..constants import date_datatypes


def convert_dates_from_model(
    source_df: pd.DataFrame,
    model_cls,
    normalize: bool | None = None,
    localize: bool | None = None,
) -> pd.DataFrame:
    """Convert django datetime columns to pandas datetime64 columns.

    Warning: When localizing, assumes stored values are tzinfo=UTC and only
    localizes dtypes datetime64[ns, UTC].
    """
    date_cols = []
    for field_cls in model_cls._meta.get_fields():
        if field_cls.get_internal_type() in date_datatypes:
            if field_cls.name in source_df.columns:
                date_cols.append(field_cls.name)
    if date_cols:
        source_df[date_cols] = source_df[date_cols].apply(pd.to_datetime, errors="coerce")
        if normalize:
            source_df[date_cols] = source_df[date_cols].apply(lambda x: x.dt.normalize())
        if localize:
            source_df[date_cols] = source_df[date_cols].apply(
                lambda x: (x.dt.tz_localize(None) if x.dtype == "datetime64[ns, UTC]" else x)
            )
    return source_df



from uuid import uuid4

from django.db import models
from django.db.models import Index
from django.utils import timezone

from edc_model.models import BaseUuidModel, HistoricalRecords


class FileHistoryManager(models.Manager):
    def get_by_natural_key(self, history_uuid):
        return self.get(history_uuid=history_uuid)


class FileHistory(BaseUuidModel):
    history_uuid = models.UUIDField(
        editable=False,
        default=uuid4,
        unique=True,
        help_text="system field for export tracking.",
    )

    model = models.CharField(max_length=50)

    export_uuid_list = models.TextField(
        default="", help_text="list of export_uuid's of model app_label.model_name"
    )

    pk_list = models.TextField(
        default="", help_text="list of pk's of model app_label.model_name"
    )

    exit_message = models.CharField(
        max_length=250, help_text="exit message from the export_transactions command"
    )

    exit_status = models.IntegerField(
        null=True, help_text="0=success, 1=failed from the export_transactions command"
    )

    filename = models.CharField(max_length=250, help_text="original filename on export")

    file_contents = models.TextField(
        default="", help_text="save contents of file as a list of rows"
    )

    exported = models.BooleanField(default=False, help_text="exported to a file")

    exported_datetime = models.DateTimeField(null=True)

    notification_plan_name = models.CharField(max_length=50, default="")

    sent = models.BooleanField(default=False, help_text="export file sent to recipient")

    sent_datetime = models.DateTimeField(null=True)

    received = models.BooleanField(
        default=False, help_text="have received an ACK from the recipient"
    )

    received_datetime = models.DateTimeField(null=True)

    closed = models.BooleanField(default=False, help_text="exported, sent, received")

    closed_datetime = models.DateTimeField(null=True)

    objects = FileHistoryManager()

    history = HistoricalRecords()

    def save(self, *args, **kwargs):
        if self.sent and self.received and self.exported and not self.closed:
            self.closed = True
            self.closed_datetime = timezone.now()
        super().save(*args, **kwargs)

    def natural_key(self):
        return (self.history_uuid,)

    class Meta(BaseUuidModel.Meta):
        indexes = (Index(fields=["sent_datetime"]),)



from io import StringIO

from django.contrib import messages
from django.http import FileResponse
from django_crypto_fields.utils import has_encrypted_fields


def export_to_csv(modeladmin, request, queryset):
    """This should be used carefully to ensure confidential
    data is not exported.
    """
    if queryset.count() == 0:
        messages.info(request, "Nothing to do.")
    elif has_encrypted_fields(queryset.model):
        messages.warning(request, "Unable to export. Report may contain confidential data")
    else:
        filename = f"{modeladmin.model._meta.label_lower.replace('.', '_')}.csv"
        buffer = StringIO()
        queryset.to_dataframe().to_csv(buffer, index=False)
        buffer.seek(0)
        response = FileResponse(buffer.read(), as_attachment=True, content_type="text/csv")
        response["Content-Disposition"] = f'filename="{filename}"'
        title = getattr(modeladmin, "change_list_title", modeladmin.model._meta.verbose_name)
        messages.success(request, f"Exported {title} to CSV. Check your browser's downloads")
        return response
    return None



from django.db.models import CharField
from django.utils.translation import gettext as _

from edc_constants.choices import IDENTITY_TYPE


class IdentityTypeField(CharField):
    """Identity type field class.

    Pair with the `Identity` field.

    Have `IdentityTypeField` immediately follow an `identity` field:

    For example:
        ...
        identity = models.CharField(
            verbose_name=_("Identity number (OMANG, etc)"),
            max_length=25,
            unique=True,
            help_text=_(
                "Use Omang, Passport number, driver's license "
                "number or Omang receipt number")
            )

        identity_type = IdentityTypeField()
        ...

    Use the value of `identity_type` to check the cleaned
    value of identity at the form level (if applicable).
    """

    description = _("Custom field for Identity Type")

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("verbose_name", _("What type of identity number is this?"))
        kwargs.setdefault("editable", True)
        kwargs.setdefault("max_length", 15)
        kwargs.setdefault("choices", IDENTITY_TYPE)
        CharField.__init__(self, *args, **kwargs)



from django.contrib import messages


class JobResult:
    def __init__(self, name=None, job_ids=None, copies=None, printer=None):
        self.name = name or ""
        self.job_ids = job_ids
        if self.job_ids:
            self.print_count = len(self.job_ids)
        else:
            self.print_count = 0
        self.copies = copies
        try:
            self.printer_info = printer.printer_info
        except AttributeError:
            self.printer_info = None
        self.jobid = ",".join([str(i) for i in self.job_ids or [0]])

    @property
    def message(self):
        return (
            f"Sent {self.print_count}/{self.copies} {self.name} labels "
            f"to printer '{self.printer_info}'. {self.job_ids}"
        )


def add_job_results_to_messages(request=None, job_results=None):
    job_results = [job for job in job_results if job]
    if job_results:
        messages.success(
            request,
            JobResult(
                name=job_results[0].name,
                job_ids=sum([j.job_ids for j in job_results], []),
                copies=sum([j.print_count for j in job_results]),
                printer=job_results[0].printer_info,
            ).message,
        )



from django.db import models
from django.db.models.deletion import PROTECT

from edc_constants.choices import YES_NO


class AeInitialSaeModelMixin(models.Model):
    sae = models.CharField(
        verbose_name="Is this event a SAE?",
        max_length=5,
        choices=YES_NO,
        help_text=(
            "(i.e. results in death, in-patient "
            "hospitalisation/prolongation, significant disability or is "
            "life-threatening)"
        ),
    )

    sae_reason = models.ForeignKey(
        "edc_adverse_event.saereason",
        on_delete=PROTECT,
        verbose_name='If "Yes", reason for SAE:',
        help_text="If subject deceased, submit a Death Report",
        blank=False,
    )

    class Meta:
        abstract = True



import pandas as pd

from edc_constants.constants import FEMALE, MALE

from ..constants import N_ONLY, N_WITH_COL_PROP, N_WITH_ROW_PROP
from ..row import RowDefinition, RowDefinitions
from ..table import Table


class BpTable(Table):

    sys_col_name = "sys_blood_pressure_avg"
    dia_col_name = "dia_blood_pressure_avg"

    def __init__(self, main_df: pd.DataFrame = None):
        super().__init__(
            colname="",
            main_df=main_df,
            title="Blood pressure at baseline (mmHg)",
        )
        self.table_df = self.table_df.reindex(index=self.table_df.index[::-1])

    @property
    def row_definitions(self) -> RowDefinitions:
        df_tmp = self.main_df.copy()
        row_defs = RowDefinitions(reverse_rows=True)
        row0 = RowDefinition(
            title=self.title,
            label=self.default_sublabel,
            condition=(df_tmp["gender"].notna()),
            columns={
                FEMALE: (N_ONLY, 2),
                MALE: (N_ONLY, 2),
                "All": (N_ONLY, 2),
            },
            drop=False,
        )
        row_defs.add(row0)
        columns = {
            FEMALE: (N_WITH_COL_PROP, 2),
            MALE: (N_WITH_COL_PROP, 2),
            "All": (N_WITH_ROW_PROP, 2),
        }
        row_defs.add(
            RowDefinition(
                label="Severe hypertension (>=180/110)",
                condition=(
                    (df_tmp[self.sys_col_name] >= 180) | (df_tmp[self.dia_col_name] >= 110)
                ),
                columns=columns,
                drop=True,
            )
        )
        row_defs.add(
            RowDefinition(
                label="Hypertension (>=140/90)",
                condition=(
                    (df_tmp[self.sys_col_name] >= 140) | (df_tmp[self.dia_col_name] >= 90)
                ),
                columns=columns,
                drop=True,
            )
        )
        row_defs.add(
            RowDefinition(
                label="Pre-hypertension (<140/90)",
                condition=(
                    (df_tmp[self.sys_col_name] >= 120) | (df_tmp[self.dia_col_name] >= 80)
                ),
                columns=columns,
                drop=True,
            )
        )
        row_defs.add(
            RowDefinition(
                label="Normal (<120/80)",
                condition=(
                    (df_tmp[self.sys_col_name] >= 90) | (df_tmp[self.dia_col_name] >= 60)
                ),
                columns=columns,
                drop=True,
            )
        )
        row_defs.add(
            RowDefinition(
                label="Low (<90/60)",
                condition=(
                    (df_tmp[self.sys_col_name] >= 0) | (df_tmp[self.dia_col_name] >= 0)
                ),
                columns=columns,
                drop=True,
            )
        )
        row_defs.add(
            RowDefinition(
                label="not measured",
                condition=(
                    (df_tmp[self.sys_col_name].isna()) & (df_tmp[self.dia_col_name].isna())
                ),
                columns=columns,
                drop=True,
            )
        )
        return row_defs



from __future__ import annotations

import csv
import sys
from pathlib import Path
from typing import Any

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.utils import timezone
from django_pandas.io import read_frame

from edc_pdutils.constants import SYSTEM_COLUMNS
from edc_sites.site import sites

from .exceptions import RandomizationListExporterError, SubjectNotRandomization
from .site_randomizers import site_randomizers


def get_randomization_list_path() -> Path:
    return Path(
        getattr(
            settings,
            "EDC_RANDOMIZATION_LIST_PATH",
            settings.ETC_DIR,
        )
    ).expanduser()


def get_assignment_for_subject(
    subject_identifier: str,
    randomizer_name: str,
    identifier_fld: str | None = None,
) -> str:
    """Returns the assignment for a randomized subject.

    Calling this method before a subject is randomized will
    raise a SubjectNotRandomization error.
    """
    obj = get_object_for_subject(
        subject_identifier, randomizer_name, identifier_fld=identifier_fld
    )
    return obj.assignment


def get_assignment_description_for_subject(
    subject_identifier: str,
    randomizer_name: str,
    identifier_fld: str | None = None,
) -> str:
    """Returns the assignment description for a randomized subject.

    Calling this method before a subject is randomized will
    raise a SubjectNotRandomization error.
    """
    randomizer_cls = site_randomizers.get(randomizer_name)
    return randomizer_cls.assignment_description_map.get(
        get_assignment_for_subject(
            subject_identifier, randomizer_name, identifier_fld=identifier_fld
        )
    )


def get_object_for_subject(
    subject_identifier: str,
    randomizer_name: str,
    identifier_fld: str | None = None,
    label: str | None = None,
) -> Any:
    """Returns a randomization list model instance or raises
    for the given subject.

    Calling this method before a subject is randomized will
    raise a SubjectNotRandomization error.
    """
    identifier_fld = identifier_fld or "subject_identifier"
    label = label or "subject"
    randomizer_cls = site_randomizers.get(randomizer_name)
    opts = {
        identifier_fld: subject_identifier,
        "randomizer_name": randomizer_name,
        "allocated": True,
        "allocated_datetime__isnull": False,
    }
    try:
        obj = randomizer_cls.model_cls().objects.get(**opts)
    except ObjectDoesNotExist as e:
        raise SubjectNotRandomization(
            f"{label.title()} not randomized. See Randomizer `{randomizer_name}`. "
            f"Got {identifier_fld}=`{subject_identifier}`."
        ) from e
    return obj


def generate_fake_randomization_list(
    all_sites=None,
    country=None,
    site_name=None,
    assignment: list | None = None,
    slots: int | None = None,
    write_header: bool | None = None,
    filename=None,
    assignment_map=None,
):
    """
    Generate a dummy randomization list.

    This trial is randomized by site so all assignments are
    the same within a site. Use this util to generate a dummy
    randomization_list.csv for import into the RandomizationList
    model. Patient registration always refers to and updates the
    RandomizationList model.

    Add slots to a dummy `randomization` list file where all
    assignments are the same for each slot.
    """
    slots = slots or 10
    assignment_map = assignment_map or ["intervention", "control"]
    if assignment not in assignment_map:
        raise ValueError(f"Invalid assignment. Got {assignment}")

    # get site ID and write the file
    site_id = sites.get_by_attr("name", site_name)
    with Path(filename).open("a+", newline="") as f:
        writer = csv.DictWriter(f, fieldnames=["sid", "assignment", "site_name", "country"])
        if write_header:
            writer.writeheader()
        for j in range(1, int(slots)):
            sid = str(j).zfill(len(str(slots)))
            writer.writerow(
                dict(
                    sid=f"{site_id}{sid}",
                    assignment=assignment,
                    site_name=site_name,
                    country=country,
                )
            )

    sys.stdout.write(f"(*) Added {slots} slots for {site_name}.\n")


def export_randomization_list(
    randomizer_name: str, path: str | None = None, username: str | None = None
) -> Path:
    randomizer_cls = site_randomizers.get(randomizer_name)

    try:
        user = get_user_model().objects.get(username=username)
    except ObjectDoesNotExist as e:
        raise RandomizationListExporterError(f"User `{username}` does not exist") from e
    if not user.has_perm(randomizer_cls.model_cls()._meta.label_lower.replace(".", ".view_")):
        raise RandomizationListExporterError(
            f"User `{username}` does not have "
            f"permission to view '{randomizer_cls.model_cls()._meta.label_lower}'"
        )
    path = Path(path or settings.EXPORT_FOLDER)
    timestamp = timezone.now().strftime("%Y%m%d%H%M")
    filename = Path(
        f"~/{settings.APP_NAME}_{randomizer_cls.name}_"
        f"randomizationlist_exported_{timestamp}.csv"
    ).expanduser()
    filename = path / filename

    df = (
        read_frame(randomizer_cls.model_cls().objects.all(), verbose=False)
        .drop(columns=SYSTEM_COLUMNS)
        .sort_values(["sid"])
        .reset_index(drop=True)
    )

    opts = dict(
        path_or_buf=filename,
        encoding="utf-8",
        index=0,
        sep="|",
    )
    df.to_csv(**opts)
    sys.stdout.write(f"{filename!s}\n")
    return Path(filename)



# Generated by Django 4.2.9 on 2024-01-30 01:00

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_auth", "0031_alter_edcpermissions_options_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="userprofile",
            name="is_multisite_viewer",
            field=models.BooleanField(default=False),
        ),
    ]



from typing import Any

from django.urls.base import reverse
from django.urls.exceptions import NoReverseMatch

from edc_dashboard.url_names import url_names


class SearchFormViewError(Exception):
    pass


class SearchFormViewMixin:
    search_form_url = None

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        kwargs.update(search_form_url_reversed=self.search_form_url_reversed)
        return super().get_context_data(**kwargs)

    @property
    def search_form_url_reversed(self):
        """Returns the reversed url selected from the url_names
        using self.search_form_url.
        """
        try:
            url = reverse(
                url_names.get(self.search_form_url), kwargs=self.search_form_url_kwargs
            )
        except NoReverseMatch as e:
            raise SearchFormViewError(
                f"{e}. Expected one of {url_names.registry}. See attribute 'search_form_url'."
            ) from e
        return f"{url}{self.querystring}"

    @property
    def search_form_url_kwargs(self):
        """Override to add custom kwargs to reverse the search form url."""
        return self.url_kwargs



from .exceptions import FormRunnerRegisterError
from .form_runner import FormRunner
from .site import site_form_runners


def register(**kwargs):
    """Registers a custom form runner."""

    def _wrapper(form_runner_cls):
        if not issubclass(form_runner_cls, (FormRunner,)):
            raise FormRunnerRegisterError(
                f"Wrapped class must a FormRunner class. Got {form_runner_cls}."
            )
        site_form_runners.register(runner=form_runner_cls)
        return form_runner_cls

    return _wrapper



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps

if TYPE_CHECKING:
    from ..models import GradingData

__all__ = ["grading_data_model_cls"]


def grading_data_model_cls() -> type[GradingData]:
    return django_apps.get_model("edc_reportable.gradingdata")



from .crf_inline_model_mixin import CrfInlineModelMixin
from .crf_model_mixin import CrfModelMixin
from .crf_no_manager_model_mixin import CrfNoManagerModelMixin
from .crf_status_model_mixin import CrfStatusModelMixin
from .crf_with_action_model_mixin import CrfWithActionModelMixin
from .singleton_crf_model_mixin import SingletonCrfModelMixin



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "edc_protocol_incident"
    verbose_name = "Edc Protocol Incidents"
    has_exportable_data = True
    include_in_administration_section = True



# Generated by Django 2.2.3 on 2019-09-09 18:59

import django.contrib.sites.models
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [("sites", "0002_alter_domain_unique")]

    operations = [
        migrations.CreateModel(
            name="EdcSite",
            fields=[],
            options={"proxy": True, "indexes": [], "constraints": []},
            bases=("sites.site",),
            managers=[("objects", django.contrib.sites.models.SiteManager())],
        ),
        migrations.CreateModel(
            name="SiteProfile",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=50, null=True)),
                ("description", models.TextField(null=True)),
                (
                    "site",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.PROTECT, to="sites.Site"
                    ),
                ),
            ],
        ),
    ]



from .action_item_admin import DataManagerActionItemAdmin
from .autocomplete_admin import (
    DataManagerUserAdmin,
    QuerySubjectAdmin,
    QueryUserAdmin,
    QueryVisitScheduleAdmin,
)
from .data_dictionary_admin import DataDictionaryAdmin
from .data_query_admin import DataQueryAdmin
from .issue_admin import IssueAdmin
from .query_rule_admin import QueryRuleAdmin



from __future__ import annotations

from dataclasses import dataclass
from uuid import UUID

__all__ = ["SubjectConsentDashboardButton"]

from typing import TYPE_CHECKING, TypeVar

from edc_view_utils import ModelButton

if TYPE_CHECKING:
    from edc_appointment.models import Appointment
    from edc_consent.model_mixins import ConsentModelMixin

    ConsentModel = TypeVar("ConsentModel", bound=ConsentModelMixin)


@dataclass
class SubjectConsentDashboardButton(ModelButton):
    """For the subject dashboard"""

    model_obj: ConsentModel = None
    model_cls: type[ConsentModel] = None
    appointment: Appointment = None

    def __post_init__(self):
        self.model_cls = self.model_obj.__class__
        if not self.next_url_name:
            self.next_url_name = "subject_dashboard_url"

    @property
    def color(self) -> str:
        return "success"

    @property
    def label(self) -> str:
        return ""

    @property
    def reverse_kwargs(self) -> dict[str, str | UUID]:
        kwargs = dict(subject_identifier=self.model_obj.subject_identifier)
        if self.appointment:
            kwargs.update(appointment=self.appointment.id)
        return kwargs



import numpy as np
import pandas as pd

try:
    from django_crypto_fields.field_cryptor import FieldCryptor
except ModuleNotFoundError:
    pass


class DecryptError(Exception):
    pass


def decrypt(row, column_name, algorithm, access_mode):
    value = np.nan
    if pd.notnull(row[column_name]):
        field_cryptor = FieldCryptor(algorithm, access_mode)
        value = field_cryptor.decrypt(row[column_name])
        if value.startswith("enc1::"):
            raise DecryptError(
                f"Failed to decrypt column value {column_name}. "
                f"Perhaps check the path to the encryption keys."
            )
    return value



from __future__ import annotations

import contextlib
import sys
from pathlib import Path

import pandas as pd
from django.apps import apps as django_apps
from django.conf import settings
from django.core.management.color import color_style
from django.db.models import QuerySet
from django.utils import timezone

from edc_export.exceptions import (
    ExporterExportFolder,
    ExporterFileExists,
    ExporterInvalidExportFormat,
)
from edc_export.utils import get_base_dir

from ..site_values_mappings import site_values_mappings
from ..utils import get_export_folder, get_model_from_table_name

style = color_style()


class Exported:
    def __init__(self, path: Path, model_name: str, record_count: int):
        self.path = path
        self.model_name = model_name
        self.record_count = record_count

    def __repr__(self):
        return f"{self.__class__.__name__}(model_name={self.model_name})"

    def __str__(self):
        return f"{self.model_name} {self.record_count}"


class CsvExporter:
    date_format: str | None = None
    delimiter: str = "|"
    encoding: str = "utf-8"
    file_exists_ok: bool = False
    index: bool = False
    sort_by: list | tuple | str | None = None

    def __init__(
        self,
        model_name: str | None = None,
        table_name: str | None = None,
        data_label: str | None = None,
        app_label: str | None = None,
        sort_by: list | tuple | str | None = None,
        export_folder: Path | None = None,
        delimiter: str | None = None,
        date_format: str | None = None,
        index: bool | None = None,
        use_simple_filename: bool | None = None,
        verbose=None,
    ):
        self.model_cls = None
        self.model_name = model_name
        self.app_label = app_label
        self.table_name = table_name
        self.delimiter = delimiter or self.delimiter
        self.date_format = date_format or self.date_format
        self.index = index or self.index
        self.sort_by = sort_by or self.sort_by
        self.use_simple_filename = use_simple_filename
        if self.sort_by and not isinstance(self.sort_by, (list, tuple)):
            self.sort_by = [self.sort_by]
        self.export_folder = export_folder or get_export_folder()
        self.verbose = verbose
        if not self.export_folder:
            raise ExporterExportFolder("Invalid export folder. Got None")
        if not self.export_folder.exists():
            raise ExporterExportFolder(f"Invalid export folder. Got {self.export_folder}")
        if self.model_name:
            with contextlib.suppress(LookupError, AttributeError):
                self.model_cls = django_apps.get_model(self.model_name)
        elif self.table_name:
            self.model_cls = get_model_from_table_name(table_name)
            if self.model_cls:
                self.model_name = self.model_cls._meta.label_lower
        self.data_label = data_label or model_name or table_name

    def to_format(
        self, export_format: str, dataframe: pd.DataFrame, export_folder: Path, **kwargs
    ) -> Exported:
        """Returns the full path of the written CSV file if the
        dataframe is exported otherwise None.

        Note: You could also just do:

            dataframe.to_csv(path_or_buf=path, **self.csv_options)

            to suppress stdout messages.
        """
        path = None
        record_count = 0
        if self.verbose:
            sys.stdout.write(self.model_name + "\r")
        if export_folder:
            self.export_folder = export_folder
        if not dataframe.empty:
            path = self.get_path()
            if self.sort_by:
                dataframe = dataframe.sort_values(by=self.sort_by)
            if self.verbose:
                sys.stdout.write(f"( ) {self.model_name} ...     \r")
            if export_format == "csv":
                path = path.with_suffix(".csv")
                dataframe.to_csv(path_or_buf=path, **self.csv_options)
            elif export_format == "stata":
                path = path.with_suffix(".dta")
                dta_version: str | None = kwargs.pop("dta_version", None)
                dta_version = int(dta_version) if dta_version else None
                dataframe.to_stata(
                    path=str(path), **self.stata_options, version=dta_version, **kwargs
                )
            else:
                raise ExporterInvalidExportFormat(
                    f"Invalid export format. Got {export_format}"
                )
            record_count = len(dataframe)
            if self.verbose:
                sys.stdout.write(
                    f"({style.SUCCESS('*')}) {self.model_name} {record_count}       \n"
                )
        elif self.verbose:
            sys.stdout.write(f"(?) {self.model_name} empty  \n")
        return Exported(path, self.model_name, record_count)

    def to_csv(self, dataframe: pd.DataFrame, export_folder: Path) -> Exported:
        """Returns the full path of the written CSV file if the
        dataframe is exported otherwise None.

        Note: You could also just do:

            dataframe.to_csv(path_or_buf=path, **self.csv_options)

            to suppress stdout messages.
        """
        return self.to_format("csv", dataframe=dataframe, export_folder=export_folder)

    def to_stata(
        self,
        dataframe: pd.DataFrame,
        export_folder: Path,
        dta_version: str | None = None,
    ) -> Exported:
        """Returns the full path of the written STATA file if the
        dataframe is exported otherwise None.
        """
        # TODO: if exporting to stata and version is <117, truncate str columns to 244
        # if dta_version and int(dta_version) < 117:
        #     dataframe = dataframe.apply(
        #         lambda x: x.apply(lambda y: y[:244] if isinstance(y, str) else y)
        #     )
        opts = dict(
            dataframe=dataframe,
            export_folder=export_folder,
            variable_labels=self.stata_variable_labels(dataframe),
            dta_version=dta_version,
        )
        return self.to_format("stata", **opts)

    @property
    def csv_options(self) -> dict:
        """Returns default options for dataframe.to_csv()."""
        return dict(
            index=self.index,
            encoding=self.encoding,
            sep=self.delimiter,
            date_format=self.date_format,
        )

    @property
    def stata_options(self) -> dict:
        """Returns default options for dataframe.to_stata()."""
        return dict(data_label=f"{self.data_label}.dta")

    def get_path(self) -> Path:
        """Returns a full path with filename."""
        root_dir = self.export_folder
        if not root_dir.exists():
            raise ExporterExportFolder(f"Base folder does not exist. Got {root_dir}.")
        base_dir = get_base_dir()
        if not (root_dir / base_dir).exists():
            (root_dir / base_dir).mkdir(parents=True, exist_ok=False)
        path = root_dir / base_dir / self.filename
        if path.exists() and not self.file_exists_ok:
            raise ExporterFileExists(
                f"File '{path}' exists! Not exporting {self.model_name}.\n"
            )
        return path

    @property
    def filename(self) -> str:
        """Returns a filename based on the timestamp."""
        if self.use_simple_filename:
            filename = self.model_name.split(".")[1].upper()
        else:
            try:
                timestamp_format = settings.EXPORT_FILENAME_TIMESTAMP_FORMAT
            except AttributeError:
                timestamp_format = "%Y%m%d%H%M%S"
            if not timestamp_format:
                suffix = ""
            else:
                suffix = f"_{timezone.now().strftime(timestamp_format)}"
            prefix = (self.model_name or self.data_label).replace("-", "_").replace(".", "_")
            filename = f"{prefix}{suffix}"
        return filename

    def stata_variable_labels(self, dataframe: pd.DataFrame) -> dict[str, str]:
        variable_labels = dict(id="primary key")
        variable_labels.update(
            {obj.field_name: obj.prompt[:79] for obj in self.data_dictionary_qs(dataframe)}
        )
        return variable_labels

    def stata_value_labels(self, dataframe: pd.DataFrame) -> list:
        commands = []
        choices = {}
        if self.model_cls:
            for field_cls in self.model_cls._meta.get_fields():
                if field_cls.get_internal_type() == "CharField" and field_cls.choices:
                    responses = []
                    for tpl in field_cls.choices:
                        if mapped_choice := site_values_mappings.get_by_choices(tpl):
                            responses.append([mapped_choice[0], mapped_choice[1]])
                        else:
                            responses.append([tpl[0], tpl[1]])
                    choices.update({field_cls.name: responses})
            for fname, responses in choices.items():
                labels = []
                if fname in list(dataframe.columns):
                    for stored, displayed in responses:
                        labels.append(f'"{stored}" "{displayed}"')
                    commands.append(f"label define {fname}l {' '.join(labels)}")
                    commands.append(f"encode {fname}, generate({fname}_encoded) {fname}l")
                    commands.append(f"ren {fname} {fname}_edc")
                    commands.append(f"ren {fname}_encoded {fname}")
                    commands.append("")
            with (self.get_path() / ".do").open("w") as f:
                f.writelines(f"{command}\n" for command in commands)
        return commands

    def data_dictionary_qs(self, dataframe: pd.DataFrame) -> QuerySet:
        data_dictionary_model_cls = django_apps.get_model("edc_data_manager.DataDictionary")
        return data_dictionary_model_cls.objects.filter(
            model=self.model_name, field_name__in=list(dataframe.columns)
        )


class Exporter(CsvExporter):
    pass



from django import forms

from edc_randomization.site_randomizers import site_randomizers

from ...models import Rx


def get_last_weight():
    return 100


class RxForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.fields.get("randomizer_name"):
            self.fields["randomizer_name"] = forms.ChoiceField(
                label="Randomizer name",
                choices=site_randomizers.get_as_choices(),
                required=False,
            )

    subject_identifier = forms.CharField(
        label="Subject Identifier",
        required=False,
        widget=forms.TextInput(attrs={"readonly": "readonly"}),
    )

    class Meta:
        model = Rx
        fields = "__all__"



from __future__ import annotations

from collections.abc import Generator


class ChoicesError(Exception):
    pass


class Choices:
    STORE = 0
    DISPLAY = 1

    def __init__(
        self,
        *args: tuple[str, str] | tuple[str, str, str | int],
        fillmeta: bool | None = None,
    ):
        meta_seen = []
        self.choices = []
        for index, arg in enumerate(args):
            store, display, meta = None, None, None
            try:
                store, display, meta = arg
            except ValueError:
                store, display = arg
            if fillmeta:
                meta = str(index + 1) if meta is None else str(meta)
                if meta not in meta_seen:
                    meta_seen.append(meta)
                else:
                    raise ChoicesError(
                        f"Meta values will not be unique. Got ({store}, {display}, {meta}). "
                        f"You may need to declare all meta values explicitly. See {args}."
                    )
            choice = (str(store), str(display), meta)
            self.choices.append(choice)
        self.choices: tuple[tuple[str, str, str], ...] = tuple(self.choices)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.choices})"

    def __str__(self) -> str:
        return str(self.choices)

    def __call__(self, *args, **kwargs) -> tuple[tuple[str, str], ...]:  # noqa: ARG002
        return tuple((c[self.STORE], c[self.DISPLAY]) for c in self.choices)

    def __iter__(self) -> Generator[tuple[str, str, str]]:
        yield from self.choices



from django.db.models.signals import m2m_changed
from django.dispatch import receiver
from simple_history.signals import post_create_historical_record

from ..site_notifications import site_notifications
from ..update_mailing_lists_in_m2m import update_mailing_lists_in_m2m


@receiver(
    post_create_historical_record,
    weak=False,
    dispatch_uid="notification_on_post_create_historical_record",
)
def notification_on_post_create_historical_record(
    instance, history_date, history_user, history_change_reason, **kwargs
):
    """Checks and processes any notifications for this model.

    Processes if `label_lower` is in site_notifications.models.

    Note, this is the post_create of the historical model.
    """
    if site_notifications.loaded and instance._meta.label_lower in site_notifications.models:
        opts = dict(
            instance=instance,
            user=instance.user_modified or instance.user_created,
            history_date=history_date,
            history_user=history_user,
            history_change_reason=history_change_reason,
            fail_silently=True,
            **kwargs,
        )
        site_notifications.notify(**opts)


@receiver(m2m_changed, weak=False, dispatch_uid="manage_mailists_on_userprofile_m2m_changed")
def manage_mailists_on_userprofile_m2m_changed(action, instance, pk_set, sender, **kwargs):
    """Updates the mail server mailing lists based on the
    selections in the UserProfile model.
    """
    try:
        instance.email_notifications
    except AttributeError:
        pass
    else:
        if action == "post_remove":
            update_mailing_lists_in_m2m(
                sender=sender,
                userprofile=instance,
                unsubscribe=True,
                pk_set=pk_set,
                verbose=True,
            )
        elif action == "post_add":
            update_mailing_lists_in_m2m(
                sender=sender,
                userprofile=instance,
                subscribe=True,
                pk_set=pk_set,
                verbose=True,
            )



from django.utils.safestring import mark_safe

CELLS_PER_MICROLITER = "cells/μL"
CELLS_PER_MILLIMETER_CUBED = "cells/mm^3"
CELLS_PER_MILLIMETER_CUBED_DISPLAY = mark_safe("cells/mm<sup>3</sup>")  # nosec B308
COPIES_PER_MILLILITER = "copies/mL"
EGFR_UNITS = "mL/min/1.73m2"
FEMTOLITERS_PER_CELL = "fL/cell"
GRAMS_PER_DECILITER = "g/dL"
GRAMS_PER_LITER = "g/L"
IU_LITER = "IU/L"
IU_LITER_DISPLAY = mark_safe("IU/L")  # nosec B308
MILLI_IU_LITER = "mIU/L"
MILLI_IU_LITER_DISPLAY = mark_safe("mIU/L")  # nosec B308
MICRO_IU_MILLILITER = "uIU/mL"
MICRO_IU_MILLILITER_DISPLAY = mark_safe("μIU/mL")  # nosec B308
MICROMOLES_PER_LITER = "umol/L"
MICROMOLES_PER_LITER_DISPLAY = "μmol/L (micromoles/L)"
MILLIGRAMS_PER_DECILITER = "mg/dL"
MILLIGRAMS_PER_LITER = "mg/L"
MILLILITER_PER_MINUTE = "mL/min"
MILLIMOLES_PER_LITER = "mmol/L"
MILLIMOLES_PER_LITER_DISPLAY = "mmol/L (millimoles/L)"
MM3 = "mm3"
MM3_DISPLAY = mark_safe("mm<sup>3</sup>")  # nosec B308
PERCENT = "%"
PICOGRAMS_PER_CELL = "pg/cell"
PLUS = "+"
TEN_X_3_PER_LITER = "10^3/L"
TEN_X_3_PER_LITER_DISPLAY = mark_safe("10<sup>3</sup>/L")  # nosec B308
TEN_X_9_PER_LITER = "10^9/L"
TEN_X_9_PER_LITER_DISPLAY = mark_safe("10<sup>9</sup>/L")  # nosec B308



# Generated by Django 2.1 on 2018-09-22 16:43

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [("edc_export", "0001_initial")]

    operations = [
        migrations.AlterModelOptions(
            name="datarequesthistory",
            options={
                "ordering": ("-exported_datetime",),
                "verbose_name": "Data Request History",
                "verbose_name_plural": "Data Request History",
            },
        )
    ]
